@@ -2,14 +2,16 @@
 Copyright (c) 2018-2020 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
-are made available under the terms of the Eclipse Public License v1.0
+are made available under the terms of the Eclipse Public License 2.0
 and Eclipse Distribution License v1.0 which accompany this distribution.
  
 The Eclipse Public License is available at
-   http://www.eclipse.org/legal/epl-v10.html
+   https://www.eclipse.org/legal/epl-2.0/
 and the Eclipse Distribution License is available at
   http://www.eclipse.org/org/documents/edl-v10.php.
  
+SPDX-License-Identifier: EPL-2.0 OR EDL-1.0
+
 Contributors:
    Roger Light - initial implementation and documentation.
 */
@@ -21,7 +23,7 @@ Contributors:
 
 struct mqtt__string {
 	char *v;
-	int len;
+	uint16_t len;
 };
 
 struct mqtt5__property {
@@ -44,7 +46,9 @@ int property__read_all(int command, struct mosquitto__packet *packet, mosquitto_
 int property__write_all(struct mosquitto__packet *packet, const mosquitto_property *property, bool write_len);
 void property__free(mosquitto_property **property);
 
-int property__get_length(const mosquitto_property *property);
-int property__get_length_all(const mosquitto_property *property);
+unsigned int property__get_length(const mosquitto_property *property);
+unsigned int property__get_length_all(const mosquitto_property *property);
+
+unsigned int property__get_remaining_length(const mosquitto_property *props);
 
 #endif
