@@ -2,14 +2,16 @@
 Copyright (c) 2009-2020 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
-are made available under the terms of the Eclipse Public License v1.0
+are made available under the terms of the Eclipse Public License 2.0
 and Eclipse Distribution License v1.0 which accompany this distribution.
  
 The Eclipse Public License is available at
-   http://www.eclipse.org/legal/epl-v10.html
+   https://www.eclipse.org/legal/epl-2.0/
 and the Eclipse Distribution License is available at
   http://www.eclipse.org/org/documents/edl-v10.php.
  
+SPDX-License-Identifier: EPL-2.0 OR EDL-1.0
+
 Contributors:
    Roger Light - initial implementation and documentation.
    Tatsuzo Osawa - Add epoll.
@@ -23,38 +25,16 @@ Contributors:
 
 #ifdef WITH_WEBSOCKETS
 #  include <libwebsockets.h>
-
-#  if defined(LWS_LIBRARY_VERSION_NUMBER)
-#    define libwebsocket_callback_on_writable(A, B) lws_callback_on_writable((B))
-#    define libwebsocket_service(A, B) lws_service((A), (B))
-#    define libwebsocket_create_context(A) lws_create_context((A))
-#    define libwebsocket_context_destroy(A) lws_context_destroy((A))
-#    define libwebsocket_write(A, B, C, D) lws_write((A), (B), (C), (D))
-#    define libwebsocket_get_socket_fd(A) lws_get_socket_fd((A))
-#    define libwebsockets_return_http_status(A, B, C, D) lws_return_http_status((B), (C), (D))
-#    define libwebsockets_get_protocol(A) lws_get_protocol((A))
-#    define libwebsocket_context lws_context
-#    define libwebsocket_protocols lws_protocols
-#    define libwebsocket_callback_reasons lws_callback_reasons
-#    define libwebsocket lws
-#  else
-#    define lws_pollfd pollfd
-#    define lws_service_fd(A, B) libwebsocket_service_fd((A), (B))
-#    define lws_pollargs libwebsocket_pollargs
-#  endif
 #endif
 
 #include "mosquitto_internal.h"
 #include "mosquitto_broker.h"
 #include "mosquitto_plugin.h"
 #include "mosquitto.h"
+#include "password_mosq.h"
 #include "tls_mosq.h"
 #include "uthash.h"
 
-#define uhpa_malloc(size) mosquitto__malloc(size)
-#define uhpa_free(ptr) mosquitto__free(ptr)
-#include "uhpa.h"
-
 #ifndef __GNUC__
 #define __attribute__(attrib)
 #endif
@@ -66,74 +46,19 @@ Contributors:
 #define MQTT3_LOG_STDOUT 0x04
 #define MQTT3_LOG_STDERR 0x08
 #define MQTT3_LOG_TOPIC 0x10
+#define MQTT3_LOG_DLT 0x20
 #define MQTT3_LOG_ALL 0xFF
 
 #define WEBSOCKET_CLIENT -2
 
-
+#define CMD_PORT_LIMIT 10
 #define TOPIC_HIERARCHY_LIMIT 200
 
-/* ========================================
- * UHPA data types
- * ======================================== */
-
-/* See uhpa.h
- *
- * The idea here is that there is potentially a lot of wasted space (and time)
- * in malloc calls for frequent, small heap allocations. This can happen if
- * small payloads are used by clients or if individual topic elements are
- * small.
- *
- * In both cases, a struct is used that includes a void* or char* pointer to
- * point to the dynamically allocated memory used. To allocate and store a
- * single byte needs the size of the pointer (8 bytes on a 64 bit
- * architecture), the malloc overhead and the memory allocated itself (which
- * will often be larger than the memory requested, on 64 bit Linux this can be
- * a minimum of 24 bytes). To allocate and store 1 byte of heap memory we need
- * in this example 32 bytes.
- *
- * UHPA uses a union to either store data in an array, or to allocate memory on
- * the heap, depending on the size of the data being stored (this does mean
- * that the size of the data must always be known). Setting the size of the
- * array changes the point at which heap allocation starts. Using the example
- * above, this means that an array size of 32 bytes should not result in any
- * wasted space, and should be quicker as well. Certainly in the case of topic
- * elements (e.g. "bar" out of "foo/bar/baz") it is likely that an array size
- * of 32 bytes will mean that the majority of heap allocations are removed.
- *
- * You can change the size of MOSQ_PAYLOAD_UNION_SIZE and
- * MOSQ_TOPIC_ELEMENT_UNION_SIZE to change the size of the uhpa array used for
- * the payload (i.e. the published part of a message) and for topic elements
- * (e.g. "foo", "bar" or "baz" in the topic "foo/bar/baz"), and so control the
- * heap allocation threshold for these data types. You should look at your
- * application to decide what values to set, but don't set them too high
- * otherwise your overall memory usage will increase.
- *
- * You could use something like heaptrack
- * http://milianw.de/blog/heaptrack-a-heap-memory-profiler-for-linux to
- * profile heap allocations.
- *
- * I would suggest that values for MOSQ_PAYLOAD_UNION_SIZE and
- * MOSQ_TOPIC_UNION_SIZE that are equivalent to
- * sizeof(void*)+malloc_usable_size(malloc(1)) are a safe value that should
- * reduce calls to malloc without increasing memory usage at all.
- */
-#define MOSQ_PAYLOAD_UNION_SIZE 8
-typedef union {
-	void *ptr;
-	char array[MOSQ_PAYLOAD_UNION_SIZE];
-} mosquitto__payload_uhpa;
-#define UHPA_ALLOC_PAYLOAD(A) UHPA_ALLOC((A)->payload, (A)->payloadlen)
-#define UHPA_ACCESS_PAYLOAD(A) UHPA_ACCESS((A)->payload, (A)->payloadlen)
-#define UHPA_FREE_PAYLOAD(A) UHPA_FREE((A)->payload, (A)->payloadlen)
-#define UHPA_MOVE_PAYLOAD(DEST, SRC) UHPA_MOVE((DEST)->payload, (SRC)->payload, (SRC)->payloadlen)
-
-/* ========================================
- * End UHPA data types
- * ======================================== */
-
 typedef uint64_t dbid_t;
 
+typedef int (*FUNC_plugin_init_v5)(mosquitto_plugin_id_t *, void **, struct mosquitto_opt *, int);
+typedef int (*FUNC_plugin_cleanup_v5)(void *, struct mosquitto_opt *, int);
+
 typedef int (*FUNC_auth_plugin_init_v4)(void **, struct mosquitto_opt *, int);
 typedef int (*FUNC_auth_plugin_cleanup_v4)(void *, struct mosquitto_opt *, int);
 typedef int (*FUNC_auth_plugin_security_init_v4)(void *, struct mosquitto_opt *, int, bool);
@@ -170,6 +95,10 @@ struct mosquitto__auth_plugin{
 	void *lib;
 	void *user_data;
 	int (*plugin_version)(void);
+	struct mosquitto_plugin_id_t *identifier;
+
+	FUNC_plugin_init_v5 plugin_init_v5;
+	FUNC_plugin_cleanup_v5 plugin_cleanup_v5;
 
 	FUNC_auth_plugin_init_v4 plugin_init_v4;
 	FUNC_auth_plugin_cleanup_v4 plugin_cleanup_v4;
@@ -209,11 +138,34 @@ struct mosquitto__auth_plugin_config
 	struct mosquitto__auth_plugin plugin;
 };
 
+struct mosquitto__callback{
+	UT_hash_handle hh; /* For callbacks that register for e.g. a specific topic */
+	struct mosquitto__callback *next, *prev; /* For typical callbacks */
+	MOSQ_FUNC_generic_callback cb;
+	void *userdata;
+	char *data; /* e.g. topic for control event */
+};
+
+struct plugin__callbacks{
+	struct mosquitto__callback *tick;
+	struct mosquitto__callback *acl_check;
+	struct mosquitto__callback *basic_auth;
+	struct mosquitto__callback *control;
+	struct mosquitto__callback *disconnect;
+	struct mosquitto__callback *ext_auth_continue;
+	struct mosquitto__callback *ext_auth_start;
+	struct mosquitto__callback *message;
+	struct mosquitto__callback *psk_key;
+	struct mosquitto__callback *reload;
+};
+
 struct mosquitto__security_options {
 	/* Any options that get added here also need considering
 	 * in config__read() with regards whether allow_anonymous
 	 * should be disabled when these options are set.
 	 */
+	struct mosquitto__unpwd *unpwd;
+	struct mosquitto__unpwd *psk_id;
 	struct mosquitto__acl_user *acl_list;
 	struct mosquitto__acl *acl_patterns;
 	char *password_file;
@@ -224,11 +176,20 @@ struct mosquitto__security_options {
 	int8_t allow_anonymous;
 	bool allow_zero_length_clientid;
 	char *auto_id_prefix;
-	int auto_id_prefix_len;
+	uint16_t auto_id_prefix_len;
+	struct plugin__callbacks plugin_callbacks;
+	mosquitto_plugin_id_t *pid; /* For registering as a "plugin" */
 };
 
+#ifdef WITH_EPOLL
+enum struct_ident{
+	id_invalid = 0,
+	id_listener = 1,
+	id_client = 2,
+};
+#endif
+
 struct mosquitto__listener {
-	int fd;
 	uint16_t port;
 	char *host;
 	char *bind_interface;
@@ -240,7 +201,7 @@ struct mosquitto__listener {
 	enum mosquitto_protocol protocol;
 	int socket_domain;
 	bool use_username_as_clientid;
-	uint8_t maximum_qos;
+	uint8_t max_qos;
 	uint16_t max_topic_alias;
 #ifdef WITH_TLS
 	char *cafile;
@@ -250,6 +211,7 @@ struct mosquitto__listener {
 	char *tls_engine;
 	char *tls_engine_kpass_sha1;
 	char *ciphers;
+	char *ciphers_tls13;
 	char *psk_hint;
 	SSL_CTX *ssl_ctx;
 	char *crlfile;
@@ -261,15 +223,30 @@ struct mosquitto__listener {
 	enum mosquitto__keyform tls_keyform;
 #endif
 #ifdef WITH_WEBSOCKETS
-	struct libwebsocket_context *ws_context;
+	struct lws_context *ws_context;
 	char *http_dir;
-	struct libwebsocket_protocols *ws_protocol;
+	struct lws_protocols *ws_protocol;
 #endif
 	struct mosquitto__security_options security_options;
-	struct mosquitto__unpwd *unpwd;
-	struct mosquitto__unpwd *psk_id;
+#ifdef WITH_UNIX_SOCKETS
+	char *unix_socket_path;
+#endif
 };
 
+
+struct mosquitto__listener_sock{
+#ifdef WITH_EPOLL
+	/* This *must* be the first element in the struct. */
+	int ident;
+#endif
+	struct mosquitto__listener *listener;
+	mosq_sock_t sock;
+};
+
+typedef struct mosquitto_plugin_id_t{
+	struct mosquitto__listener *listener;
+} mosquitto_plugin_id_t;
+
 struct mosquitto__config {
 	bool allow_duplicate_messages;
 	int autosave_interval;
@@ -277,21 +254,28 @@ struct mosquitto__config {
 	bool check_retain_source;
 	char *clientid_prefixes;
 	bool connection_messages;
+	uint16_t cmd_port[CMD_PORT_LIMIT];
+	int cmd_port_count;
 	bool daemon;
 	struct mosquitto__listener default_listener;
 	struct mosquitto__listener *listeners;
 	int listener_count;
-	int log_dest;
+	bool local_only;
+	unsigned int log_dest;
 	int log_facility;
 	unsigned int log_type;
 	bool log_timestamp;
 	char *log_timestamp_format;
 	char *log_file;
 	FILE *log_fptr;
-	uint16_t max_inflight_messages;
-	uint16_t max_keepalive;
+	size_t max_inflight_bytes;
+	size_t max_queued_bytes;
+	int max_queued_messages;
 	uint32_t max_packet_size;
 	uint32_t message_size_limit;
+	uint16_t max_inflight_messages;
+	uint16_t max_keepalive;
+	uint8_t max_qos;
 	bool persistence;
 	char *persistence_location;
 	char *persistence_file;
@@ -307,7 +291,7 @@ struct mosquitto__config {
 	char *user;
 #ifdef WITH_WEBSOCKETS
 	int websockets_log_level;
-	int websockets_headers_size;
+	uint16_t websockets_headers_size;
 	bool have_websockets_listener;
 #endif
 #ifdef WITH_BRIDGE
@@ -346,6 +330,20 @@ struct mosquitto__subhier {
 	struct mosquitto__subhier *children;
 	struct mosquitto__subleaf *subs;
 	struct mosquitto__subshared *shared;
+	char *topic;
+	uint16_t topic_len;
+};
+
+struct sub__token {
+	struct sub__token *next;
+	char *topic;
+	uint16_t topic_len;
+};
+
+struct mosquitto__retainhier {
+	UT_hash_handle hh;
+	struct mosquitto__retainhier *parent;
+	struct mosquitto__retainhier *children;
 	struct mosquitto_msg_store *retained;
 	char *topic;
 	uint16_t topic_len;
@@ -369,14 +367,14 @@ struct mosquitto_msg_store{
 	int ref_count;
 	char* topic;
 	mosquitto_property *properties;
-	mosquitto__payload_uhpa payload;
+	void *payload;
 	time_t message_expiry_time;
 	uint32_t payloadlen;
 	uint16_t source_mid;
 	uint16_t mid;
 	uint8_t qos;
 	bool retain;
-	uint8_t origin;
+	enum mosquitto_msg_origin origin;
 };
 
 struct mosquitto_client_msg{
@@ -393,15 +391,19 @@ struct mosquitto_client_msg{
 	bool dup;
 };
 
+
 struct mosquitto__unpwd{
+	UT_hash_handle hh;
 	char *username;
 	char *password;
+	char *clientid;
 #ifdef WITH_TLS
+	unsigned char *salt;
 	unsigned int password_len;
 	unsigned int salt_len;
-	unsigned char *salt;
+	int iterations;
 #endif
-	UT_hash_handle hh;
+	enum mosquitto_pwhash_type hashtype;
 };
 
 struct mosquitto__acl{
@@ -418,11 +420,24 @@ struct mosquitto__acl_user{
 	struct mosquitto__acl *acl;
 };
 
+
+struct mosquitto_message_v5{
+	struct mosquitto_message_v5 *next, *prev;
+	char *topic;
+	void *payload;
+	mosquitto_property *properties;
+	char *clientid; /* Used only by mosquitto_broker_publish*() to indicate
+					   this message is for a specific client. */
+	int payloadlen;
+	int qos;
+	bool retain;
+};
+
+
 struct mosquitto_db{
 	dbid_t last_db_id;
 	struct mosquitto__subhier *subs;
-	struct mosquitto__unpwd *unpwd;
-	struct mosquitto__unpwd *psk_id;
+	struct mosquitto__retainhier *retains;
 	struct mosquitto *contexts_by_id;
 	struct mosquitto *contexts_by_sock;
 	struct mosquitto *contexts_for_free;
@@ -432,6 +447,8 @@ struct mosquitto_db{
 	struct clientid__index_hash *clientid_index_hash;
 	struct mosquitto_msg_store *msg_store;
 	struct mosquitto_msg_store_load *msg_store_load;
+	time_t now_s; /* Monotonic clock, where possible */
+	time_t now_real_s; /* Read clock, for measuring session/message expiry */
 #ifdef WITH_BRIDGE
 	int bridge_count;
 #endif
@@ -451,6 +468,7 @@ struct mosquitto_db{
 #ifdef WITH_EPOLL
 	int epollfd;
 #endif
+	struct mosquitto_message_v5 *plugin_msgs;
 };
 
 enum mosquitto__bridge_direction{
@@ -468,17 +486,17 @@ enum mosquitto_bridge_start_type{
 
 struct mosquitto__bridge_topic{
 	char *topic;
-	int qos;
-	enum mosquitto__bridge_direction direction;
 	char *local_prefix;
 	char *remote_prefix;
 	char *local_topic; /* topic prefixed with local_prefix */
 	char *remote_topic; /* topic prefixed with remote_prefix */
+	enum mosquitto__bridge_direction direction;
+	uint8_t qos;
 };
 
 struct bridge_address{
 	char *address;
-	int port;
+	uint16_t port;
 };
 
 struct mosquitto__bridge{
@@ -492,7 +510,8 @@ struct mosquitto__bridge{
 	bool try_private;
 	bool try_private_accepted;
 	bool clean_start;
-	int keepalive;
+	int8_t clean_start_local;
+	uint16_t keepalive;
 	struct mosquitto__bridge_topic *topics;
 	int topic_count;
 	bool topic_remapping;
@@ -505,6 +524,7 @@ struct mosquitto__bridge{
 	char *local_username;
 	char *local_password;
 	char *notification_topic;
+	char *bind_address;
 	bool notifications;
 	bool notifications_local_only;
 	enum mosquitto_bridge_start_type start_type;
@@ -513,9 +533,11 @@ struct mosquitto__bridge{
 	int backoff_base;
 	int backoff_cap;
 	int threshold;
+	uint32_t maximum_packet_size;
 	bool lazy_reconnect;
 	bool attempt_unsubscribe;
 	bool initial_notification_done;
+	bool outgoing_retain;
 #ifdef WITH_TLS
 	bool tls_insecure;
 	bool tls_ocsp_required;
@@ -544,138 +566,202 @@ struct libws_mqtt_data {
 
 #include <net_mosq.h>
 
+
+extern struct mosquitto_db db;
+
 /* ============================================================
  * Main functions
  * ============================================================ */
-int mosquitto_main_loop(struct mosquitto_db *db, mosq_sock_t *listensock, int listensock_count);
-struct mosquitto_db *mosquitto__get_db(void);
+int mosquitto_main_loop(struct mosquitto__listener_sock *listensock, int listensock_count);
 
 /* ============================================================
  * Config functions
  * ============================================================ */
 /* Initialise config struct to default values. */
-void config__init(struct mosquitto_db *db, struct mosquitto__config *config);
+void config__init(struct mosquitto__config *config);
 /* Parse command line options into config. */
-int config__parse_args(struct mosquitto_db *db, struct mosquitto__config *config, int argc, char *argv[]);
+int config__parse_args(struct mosquitto__config *config, int argc, char *argv[]);
 /* Read configuration data from config->config_file into config.
  * If reload is true, don't process config options that shouldn't be reloaded (listeners etc)
  * Returns 0 on success, 1 if there is a configuration error or if a file cannot be opened.
  */
-int config__read(struct mosquitto_db *db, struct mosquitto__config *config, bool reload);
+int config__read(struct mosquitto__config *config, bool reload);
 /* Free all config data. */
 void config__cleanup(struct mosquitto__config *config);
 int config__get_dir_files(const char *include_dir, char ***files, int *file_count);
 
-int drop_privileges(struct mosquitto__config *config, bool temporary);
-int restore_privileges(void);
+int drop_privileges(struct mosquitto__config *config);
 
 /* ============================================================
  * Server send functions
  * ============================================================ */
-int send__connack(struct mosquitto_db *db, struct mosquitto *context, int ack, int reason_code, const mosquitto_property *properties);
+int send__connack(struct mosquitto *context, uint8_t ack, uint8_t reason_code, const mosquitto_property *properties);
 int send__suback(struct mosquitto *context, uint16_t mid, uint32_t payloadlen, const void *payload);
 int send__unsuback(struct mosquitto *context, uint16_t mid, int reason_code_count, uint8_t *reason_codes, const mosquitto_property *properties);
-int send__auth(struct mosquitto_db *db, struct mosquitto *context, int reason_code, const void *auth_data, uint16_t auth_data_len);
+int send__auth(struct mosquitto *context, uint8_t reason_code, const void *auth_data, uint16_t auth_data_len);
 
 /* ============================================================
  * Network functions
  * ============================================================ */
 void net__broker_init(void);
 void net__broker_cleanup(void);
-int net__socket_accept(struct mosquitto_db *db, mosq_sock_t listensock);
+struct mosquitto *net__socket_accept(struct mosquitto__listener_sock *listensock);
 int net__socket_listen(struct mosquitto__listener *listener);
-int net__socket_get_address(mosq_sock_t sock, char *buf, int len);
+int net__socket_get_address(mosq_sock_t sock, char *buf, size_t len, uint16_t *remote_address);
 int net__tls_load_verify(struct mosquitto__listener *listener);
 int net__tls_server_ctx(struct mosquitto__listener *listener);
+int net__load_certificates(struct mosquitto__listener *listener);
 
 /* ============================================================
  * Read handling functions
  * ============================================================ */
-int handle__packet(struct mosquitto_db *db, struct mosquitto *context);
-int handle__connack(struct mosquitto_db *db, struct mosquitto *context);
-int handle__connect(struct mosquitto_db *db, struct mosquitto *context);
-int handle__disconnect(struct mosquitto_db *db, struct mosquitto *context);
-int handle__publish(struct mosquitto_db *db, struct mosquitto *context);
-int handle__subscribe(struct mosquitto_db *db, struct mosquitto *context);
-int handle__unsubscribe(struct mosquitto_db *db, struct mosquitto *context);
-int handle__auth(struct mosquitto_db *db, struct mosquitto *context);
+int handle__packet(struct mosquitto *context);
+int handle__connack(struct mosquitto *context);
+int handle__connect(struct mosquitto *context);
+int handle__disconnect(struct mosquitto *context);
+int handle__publish(struct mosquitto *context);
+int handle__subscribe(struct mosquitto *context);
+int handle__unsubscribe(struct mosquitto *context);
+int handle__auth(struct mosquitto *context);
 
 /* ============================================================
  * Database handling
  * ============================================================ */
-int db__open(struct mosquitto__config *config, struct mosquitto_db *db);
-int db__close(struct mosquitto_db *db);
+int db__open(struct mosquitto__config *config);
+int db__close(void);
 #ifdef WITH_PERSISTENCE
-int persist__backup(struct mosquitto_db *db, bool shutdown);
-int persist__restore(struct mosquitto_db *db);
+int persist__backup(bool shutdown);
+int persist__restore(void);
 #endif
-void db__limits_set(unsigned long inflight_bytes, int queued, unsigned long queued_bytes);
 /* Return the number of in-flight messages in count. */
 int db__message_count(int *count);
-int db__message_delete_outgoing(struct mosquitto_db *db, struct mosquitto *context, uint16_t mid, enum mosquitto_msg_state expect_state, int qos);
-int db__message_insert(struct mosquitto_db *db, struct mosquitto *context, uint16_t mid, enum mosquitto_msg_direction dir, int qos, bool retain, struct mosquitto_msg_store *stored, mosquitto_property *properties);
-int db__message_release_incoming(struct mosquitto_db *db, struct mosquitto *context, uint16_t mid);
+int db__message_delete_outgoing(struct mosquitto *context, uint16_t mid, enum mosquitto_msg_state expect_state, int qos);
+int db__message_insert(struct mosquitto *context, uint16_t mid, enum mosquitto_msg_direction dir, uint8_t qos, bool retain, struct mosquitto_msg_store *stored, mosquitto_property *properties, bool update);
+int db__message_release_incoming(struct mosquitto *context, uint16_t mid);
 int db__message_update_outgoing(struct mosquitto *context, uint16_t mid, enum mosquitto_msg_state state, int qos);
-int db__message_write(struct mosquitto_db *db, struct mosquitto *context);
 void db__message_dequeue_first(struct mosquitto *context, struct mosquitto_msg_data *msg_data);
-int db__messages_delete(struct mosquitto_db *db, struct mosquitto *context);
-int db__messages_easy_queue(struct mosquitto_db *db, struct mosquitto *context, const char *topic, int qos, uint32_t payloadlen, const void *payload, int retain, uint32_t message_expiry_interval, mosquitto_property **properties);
-int db__message_store(struct mosquitto_db *db, const struct mosquitto *source, uint16_t source_mid, char *topic, int qos, uint32_t payloadlen, mosquitto__payload_uhpa *payload, int retain, struct mosquitto_msg_store **stored, uint32_t message_expiry_interval, mosquitto_property *properties, dbid_t store_id, enum mosquitto_msg_origin origin);
+int db__messages_delete(struct mosquitto *context, bool force_free);
+int db__messages_easy_queue(struct mosquitto *context, const char *topic, uint8_t qos, uint32_t payloadlen, const void *payload, int retain, uint32_t message_expiry_interval, mosquitto_property **properties);
+int db__message_store(const struct mosquitto *source, struct mosquitto_msg_store *stored, uint32_t message_expiry_interval, dbid_t store_id, enum mosquitto_msg_origin origin);
 int db__message_store_find(struct mosquitto *context, uint16_t mid, struct mosquitto_msg_store **stored);
-void db__msg_store_add(struct mosquitto_db *db, struct mosquitto_msg_store *store);
-void db__msg_store_remove(struct mosquitto_db *db, struct mosquitto_msg_store *store);
+void db__msg_store_add(struct mosquitto_msg_store *store);
+void db__msg_store_remove(struct mosquitto_msg_store *store);
 void db__msg_store_ref_inc(struct mosquitto_msg_store *store);
-void db__msg_store_ref_dec(struct mosquitto_db *db, struct mosquitto_msg_store **store);
-void db__msg_store_clean(struct mosquitto_db *db);
-void db__msg_store_compact(struct mosquitto_db *db);
-int db__message_reconnect_reset(struct mosquitto_db *db, struct mosquitto *context);
-void sys_tree__init(struct mosquitto_db *db);
-void sys_tree__update(struct mosquitto_db *db, int interval, time_t start_time);
+void db__msg_store_ref_dec(struct mosquitto_msg_store **store);
+void db__msg_store_clean(void);
+void db__msg_store_compact(void);
+void db__msg_store_free(struct mosquitto_msg_store *store);
+int db__message_reconnect_reset(struct mosquitto *context);
+bool db__ready_for_flight(struct mosquitto_msg_data *msgs, int qos);
+bool db__ready_for_queue(struct mosquitto *context, int qos, struct mosquitto_msg_data *msg_data);
+void sys_tree__init(void);
+void sys_tree__update(int interval, time_t start_time);
+int db__message_write_inflight_out_all(struct mosquitto *context);
+int db__message_write_inflight_out_latest(struct mosquitto *context);
+int db__message_write_queued_out(struct mosquitto *context);
+int db__message_write_queued_in(struct mosquitto *context);
 
 /* ============================================================
  * Subscription functions
  * ============================================================ */
-int sub__add(struct mosquitto_db *db, struct mosquitto *context, const char *sub, int qos, uint32_t identifier, int options, struct mosquitto__subhier **root);
-struct mosquitto__subhier *sub__add_hier_entry(struct mosquitto__subhier *parent, struct mosquitto__subhier **sibling, const char *topic, size_t len);
-int sub__remove(struct mosquitto_db *db, struct mosquitto *context, const char *sub, struct mosquitto__subhier *root, uint8_t *reason);
+int sub__add(struct mosquitto *context, const char *sub, uint8_t qos, uint32_t identifier, int options, struct mosquitto__subhier **root);
+struct mosquitto__subhier *sub__add_hier_entry(struct mosquitto__subhier *parent, struct mosquitto__subhier **sibling, const char *topic, uint16_t len);
+int sub__remove(struct mosquitto *context, const char *sub, struct mosquitto__subhier *root, uint8_t *reason);
 void sub__tree_print(struct mosquitto__subhier *root, int level);
-int sub__clean_session(struct mosquitto_db *db, struct mosquitto *context);
-int sub__retain_queue(struct mosquitto_db *db, struct mosquitto *context, const char *sub, int sub_qos, uint32_t subscription_identifier);
-int sub__messages_queue(struct mosquitto_db *db, const char *source_id, const char *topic, int qos, int retain, struct mosquitto_msg_store **stored);
+int sub__clean_session(struct mosquitto *context);
+int sub__messages_queue(const char *source_id, const char *topic, uint8_t qos, int retain, struct mosquitto_msg_store **stored);
+int sub__topic_tokenise(const char *subtopic, char **local_sub, char ***topics, const char **sharename);
+void sub__topic_tokens_free(struct sub__token *tokens);
 
 /* ============================================================
  * Context functions
  * ============================================================ */
-struct mosquitto *context__init(struct mosquitto_db *db, mosq_sock_t sock);
-void context__cleanup(struct mosquitto_db *db, struct mosquitto *context, bool do_free);
-void context__disconnect(struct mosquitto_db *db, struct mosquitto *context);
-void context__add_to_disused(struct mosquitto_db *db, struct mosquitto *context);
-void context__free_disused(struct mosquitto_db *db);
-void context__send_will(struct mosquitto_db *db, struct mosquitto *context);
-void context__remove_from_by_id(struct mosquitto_db *db, struct mosquitto *context);
+struct mosquitto *context__init(mosq_sock_t sock);
+void context__cleanup(struct mosquitto *context, bool force_free);
+void context__disconnect(struct mosquitto *context);
+void context__add_to_disused(struct mosquitto *context);
+void context__free_disused(void);
+void context__send_will(struct mosquitto *context);
+void context__remove_from_by_id(struct mosquitto *context);
+
+int connect__on_authorised(struct mosquitto *context, void *auth_data_out, uint16_t auth_data_out_len);
+
+
+/* ============================================================
+ * Control functions
+ * ============================================================ */
+#ifdef WITH_CONTROL
+int control__process(struct mosquitto *context, struct mosquitto_msg_store *stored);
+void control__cleanup(void);
+#endif
+int control__register_callback(struct mosquitto__security_options *opts, MOSQ_FUNC_generic_callback cb_func, const char *topic, void *userdata);
+int control__unregister_callback(struct mosquitto__security_options *opts, MOSQ_FUNC_generic_callback cb_func, const char *topic);
 
-int connect__on_authorised(struct mosquitto_db *db, struct mosquitto *context, void *auth_data_out, uint16_t auth_data_out_len);
 
 /* ============================================================
  * Logging functions
  * ============================================================ */
 int log__init(struct mosquitto__config *config);
 int log__close(struct mosquitto__config *config);
-int log__printf(struct mosquitto *mosq, int level, const char *fmt, ...) __attribute__((format(printf, 3, 4)));
+int log__printf(struct mosquitto *mosq, unsigned int level, const char *fmt, ...) __attribute__((format(printf, 3, 4)));
 void log__internal(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
 
 /* ============================================================
  * Bridge functions
  * ============================================================ */
 #ifdef WITH_BRIDGE
-int bridge__new(struct mosquitto_db *db, struct mosquitto__bridge *bridge);
-int bridge__connect(struct mosquitto_db *db, struct mosquitto *context);
-int bridge__connect_step1(struct mosquitto_db *db, struct mosquitto *context);
-int bridge__connect_step2(struct mosquitto_db *db, struct mosquitto *context);
-int bridge__connect_step3(struct mosquitto_db *db, struct mosquitto *context);
+void bridge__start_all(void);
+int bridge__new(struct mosquitto__bridge *bridge);
+void bridge__cleanup(struct mosquitto *context);
+int bridge__connect(struct mosquitto *context);
+int bridge__connect_step1(struct mosquitto *context);
+int bridge__connect_step2(struct mosquitto *context);
+int bridge__connect_step3(struct mosquitto *context);
+int bridge__on_connect(struct mosquitto *context);
 void bridge__packet_cleanup(struct mosquitto *context);
+void bridge_check(void);
+int bridge__register_local_connections(void);
+int bridge__add_topic(struct mosquitto__bridge *bridge, const char *topic, enum mosquitto__bridge_direction direction, uint8_t qos, const char *local_prefix, const char *remote_prefix);
+int bridge__remap_topic_in(struct mosquitto *context, char **topic);
 #endif
 
+/* ============================================================
+ * IO multiplex related functions
+ * ============================================================ */
+int mux__init(struct mosquitto__listener_sock *listensock, int listensock_count);
+int mux__loop_prepare(void);
+int mux__add_out(struct mosquitto *context);
+int mux__remove_out(struct mosquitto *context);
+int mux__add_in(struct mosquitto *context);
+int mux__delete(struct mosquitto *context);
+int mux__wait(void);
+int mux__handle(struct mosquitto__listener_sock *listensock, int listensock_count);
+int mux__cleanup(void);
+
+/* ============================================================
+ * Listener related functions
+ * ============================================================ */
+void listener__set_defaults(struct mosquitto__listener *listener);
+void listeners__reload_all_certificates(void);
+
+/* ============================================================
+ * Plugin related functions
+ * ============================================================ */
+int plugin__load_v5(struct mosquitto__listener *listener, struct mosquitto__auth_plugin *plugin, struct mosquitto_opt *auth_options, int auth_option_count, void *lib);
+void plugin__handle_disconnect(struct mosquitto *context, int reason);
+int plugin__handle_message(struct mosquitto *context, struct mosquitto_msg_store *stored);
+void LIB_ERROR(void);
+void plugin__handle_tick(void);
+
+/* ============================================================
+ * Property related functions
+ * ============================================================ */
+int keepalive__add(struct mosquitto *context);
+void keepalive__check(void);
+int keepalive__remove(struct mosquitto *context);
+void keepalive__remove_all(void);
+int keepalive__update(struct mosquitto *context);
+
 /* ============================================================
  * Property related functions
  * ============================================================ */
@@ -683,38 +769,46 @@ int property__process_connect(struct mosquitto *context, mosquitto_property **pr
 int property__process_will(struct mosquitto *context, struct mosquitto_message_all *msg, mosquitto_property **props);
 int property__process_disconnect(struct mosquitto *context, mosquitto_property **props);
 
+/* ============================================================
+ * Retain tree related functions
+ * ============================================================ */
+int retain__init(void);
+void retain__clean(struct mosquitto__retainhier **retainhier);
+int retain__queue(struct mosquitto *context, const char *sub, uint8_t sub_qos, uint32_t subscription_identifier);
+int retain__store(const char *topic, struct mosquitto_msg_store *stored, char **split_topics);
+
 /* ============================================================
  * Security related functions
  * ============================================================ */
-int acl__find_acls(struct mosquitto_db *db, struct mosquitto *context);
-int mosquitto_security_module_init(struct mosquitto_db *db);
-int mosquitto_security_module_cleanup(struct mosquitto_db *db);
+int acl__find_acls(struct mosquitto *context);
+int mosquitto_security_module_init(void);
+int mosquitto_security_module_cleanup(void);
 
-int mosquitto_security_init(struct mosquitto_db *db, bool reload);
-int mosquitto_security_apply(struct mosquitto_db *db);
-int mosquitto_security_cleanup(struct mosquitto_db *db, bool reload);
-int mosquitto_acl_check(struct mosquitto_db *db, struct mosquitto *context, const char *topic, long payloadlen, void* payload, int qos, bool retain, int access);
-int mosquitto_unpwd_check(struct mosquitto_db *db, struct mosquitto *context, const char *username, const char *password);
-int mosquitto_psk_key_get(struct mosquitto_db *db, struct mosquitto *context, const char *hint, const char *identity, char *key, int max_key_len);
+int mosquitto_security_init(bool reload);
+int mosquitto_security_apply(void);
+int mosquitto_security_cleanup(bool reload);
+int mosquitto_acl_check(struct mosquitto *context, const char *topic, uint32_t payloadlen, void* payload, uint8_t qos, bool retain, int access);
+int mosquitto_unpwd_check(struct mosquitto *context);
+int mosquitto_psk_key_get(struct mosquitto *context, const char *hint, const char *identity, char *key, int max_key_len);
 
-int mosquitto_security_init_default(struct mosquitto_db *db, bool reload);
-int mosquitto_security_apply_default(struct mosquitto_db *db);
-int mosquitto_security_cleanup_default(struct mosquitto_db *db, bool reload);
-int mosquitto_acl_check_default(struct mosquitto_db *db, struct mosquitto *context, const char *topic, int access);
-int mosquitto_unpwd_check_default(struct mosquitto_db *db, struct mosquitto *context, const char *username, const char *password);
-int mosquitto_psk_key_get_default(struct mosquitto_db *db, struct mosquitto *context, const char *hint, const char *identity, char *key, int max_key_len);
+int mosquitto_security_init_default(bool reload);
+int mosquitto_security_apply_default(void);
+int mosquitto_security_cleanup_default(bool reload);
+int mosquitto_psk_key_get_default(struct mosquitto *context, const char *hint, const char *identity, char *key, int max_key_len);
 
-int mosquitto_security_auth_start(struct mosquitto_db *db, struct mosquitto *context, bool reauth, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len);
-int mosquitto_security_auth_continue(struct mosquitto_db *db, struct mosquitto *context, const void *data_in, uint16_t data_len, void **data_out, uint16_t *data_out_len);
+int mosquitto_security_auth_start(struct mosquitto *context, bool reauth, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len);
+int mosquitto_security_auth_continue(struct mosquitto *context, const void *data_in, uint16_t data_len, void **data_out, uint16_t *data_out_len);
+
+void unpwd__free_item(struct mosquitto__unpwd **unpwd, struct mosquitto__unpwd *item);
 
 /* ============================================================
  * Session expiry
  * ============================================================ */
-int session_expiry__add(struct mosquitto_db *db, struct mosquitto *context);
+int session_expiry__add(struct mosquitto *context);
 void session_expiry__remove(struct mosquitto *context);
-void session_expiry__remove_all(struct mosquitto_db *db);
-void session_expiry__check(struct mosquitto_db *db, time_t now);
-void session_expiry__send_all(struct mosquitto_db *db);
+void session_expiry__remove_all(void);
+void session_expiry__check(void);
+void session_expiry__send_all(void);
 
 /* ============================================================
  * Window service and signal related functions
@@ -731,21 +825,25 @@ DWORD WINAPI SigThreadProc(void* data);
  * Websockets related functions
  * ============================================================ */
 #ifdef WITH_WEBSOCKETS
-#  if defined(LWS_LIBRARY_VERSION_NUMBER)
 struct lws_context *mosq_websockets_init(struct mosquitto__listener *listener, const struct mosquitto__config *conf);
-#  else
-struct libwebsocket_context *mosq_websockets_init(struct mosquitto__listener *listener, const struct mosquitto__config *conf);
-#  endif
 #endif
-void do_disconnect(struct mosquitto_db *db, struct mosquitto *context, int reason);
+void do_disconnect(struct mosquitto *context, int reason);
 
 /* ============================================================
  * Will delay
  * ============================================================ */
 int will_delay__add(struct mosquitto *context);
-void will_delay__check(struct mosquitto_db *db, time_t now);
-void will_delay__send_all(struct mosquitto_db *db);
+void will_delay__check(void);
+void will_delay__send_all(void);
 void will_delay__remove(struct mosquitto *mosq);
 
+
+/* ============================================================
+ * Other
+ * ============================================================ */
+#ifdef WITH_XTREPORT
+void xtreport(void);
+#endif
+
 #endif
 
