@@ -2,14 +2,16 @@
 Copyright (c) 2010-2020 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
-are made available under the terms of the Eclipse Public License v1.0
+are made available under the terms of the Eclipse Public License 2.0
 and Eclipse Distribution License v1.0 which accompany this distribution.
 
 The Eclipse Public License is available at
-   http://www.eclipse.org/legal/epl-v10.html
+   https://www.eclipse.org/legal/epl-2.0/
 and the Eclipse Distribution License is available at
   http://www.eclipse.org/org/documents/edl-v10.php.
 
+SPDX-License-Identifier: EPL-2.0 OR EDL-1.0
+
 Contributors:
    Roger Light - initial implementation and documentation.
 */
@@ -41,18 +43,18 @@ uint32_t db_version;
 
 const unsigned char magic[15] = {0x00, 0xB5, 0x00, 'm','o','s','q','u','i','t','t','o',' ','d','b'};
 
-static int persist__restore_sub(struct mosquitto_db *db, const char *client_id, const char *sub, int qos, uint32_t identifier, int options);
+static int persist__restore_sub(const char *client_id, const char *sub, uint8_t qos, uint32_t identifier, int options);
 
-static struct mosquitto *persist__find_or_add_context(struct mosquitto_db *db, const char *client_id, uint16_t last_mid)
+static struct mosquitto *persist__find_or_add_context(const char *client_id, uint16_t last_mid)
 {
 	struct mosquitto *context;
 
 	if(!client_id) return NULL;
 
 	context = NULL;
-	HASH_FIND(hh_id, db->contexts_by_id, client_id, strlen(client_id), context);
+	HASH_FIND(hh_id, db.contexts_by_id, client_id, strlen(client_id), context);
 	if(!context){
-		context = context__init(db, -1);
+		context = context__init(-1);
 		if(!context) return NULL;
 		context->id = mosquitto__strdup(client_id);
 		if(!context->id){
@@ -62,7 +64,7 @@ static struct mosquitto *persist__find_or_add_context(struct mosquitto_db *db, c
 
 		context->clean_start = false;
 
-		HASH_ADD_KEYPTR(hh_id, db->contexts_by_id, context->id, strlen(context->id), context);
+		HASH_ADD_KEYPTR(hh_id, db.contexts_by_id, context->id, strlen(context->id), context);
 	}
 	if(last_mid){
 		context->last_mid = last_mid;
@@ -76,9 +78,8 @@ int persist__read_string_len(FILE *db_fptr, char **str, uint16_t len)
 	char *s = NULL;
 
 	if(len){
-		s = mosquitto__malloc(len+1);
+		s = mosquitto__malloc(len+1U);
 		if(!s){
-			fclose(db_fptr);
 			log__printf(NULL, MOSQ_LOG_ERR, "Error: Out of memory.");
 			return MOSQ_ERR_NOMEM;
 		}
@@ -108,19 +109,25 @@ int persist__read_string(FILE *db_fptr, char **str)
 }
 
 
-static int persist__client_msg_restore(struct mosquitto_db *db, struct P_client_msg *chunk)
+static int persist__client_msg_restore(struct P_client_msg *chunk)
 {
 	struct mosquitto_client_msg *cmsg;
 	struct mosquitto_msg_store_load *load;
 	struct mosquitto *context;
 	struct mosquitto_msg_data *msg_data;
 
-	HASH_FIND(hh, db->msg_store_load, &chunk->F.store_id, sizeof(dbid_t), load);
+	HASH_FIND(hh, db.msg_store_load, &chunk->F.store_id, sizeof(dbid_t), load);
 	if(!load){
 		/* Can't find message - probably expired */
 		return MOSQ_ERR_SUCCESS;
 	}
 
+	context = persist__find_or_add_context(chunk->client_id, 0);
+	if(!context){
+		log__printf(NULL, MOSQ_LOG_WARNING, "Warning: Persistence file contains client message with no matching client. File may be corrupt.");
+		return 0;
+	}
+
 	cmsg = mosquitto__calloc(1, sizeof(struct mosquitto_client_msg));
 	if(!cmsg){
 		log__printf(NULL, MOSQ_LOG_ERR, "Error: Out of memory.");
@@ -141,13 +148,6 @@ static int persist__client_msg_restore(struct mosquitto_db *db, struct P_client_
 	cmsg->store = load->store;
 	db__msg_store_ref_inc(cmsg->store);
 
-	context = persist__find_or_add_context(db, chunk->client_id, 0);
-	if(!context){
-		mosquitto__free(cmsg);
-		log__printf(NULL, MOSQ_LOG_ERR, "Error restoring persistent database, message store corrupt.");
-		return 1;
-	}
-
 	if(cmsg->direction == mosq_md_out){
 		msg_data = &context->msgs_out;
 	}else{
@@ -173,7 +173,7 @@ static int persist__client_msg_restore(struct mosquitto_db *db, struct P_client_
 }
 
 
-static int persist__client_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
+static int persist__client_chunk_restore(FILE *db_fptr)
 {
 	int i, rc = 0;
 	struct mosquitto *context;
@@ -186,12 +186,15 @@ static int persist__client_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
 	}else{
 		rc = persist__chunk_client_read_v234(db_fptr, &chunk, db_version);
 	}
-	if(rc){
-		fclose(db_fptr);
+	if(rc > 0){
 		return rc;
+	}else if(rc < 0){
+		/* Client not loaded, but otherwise not an error */
+		log__printf(NULL, MOSQ_LOG_WARNING, "Warning: Empty client entry found in persistence database, it may be corrupt.");
+		return MOSQ_ERR_SUCCESS;
 	}
 
-	context = persist__find_or_add_context(db, chunk.client_id, chunk.F.last_mid);
+	context = persist__find_or_add_context(chunk.client_id, chunk.F.last_mid);
 	if(context){
 		context->session_expiry_time = chunk.F.session_expiry_time;
 		context->session_expiry_interval = chunk.F.session_expiry_interval;
@@ -199,13 +202,13 @@ static int persist__client_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
 			/* username is not freed here, it is now owned by context */
 			context->username = chunk.username;
 			chunk.username = NULL;
-			/* in per_listener_settings mode, try to find the listener by persisted port */
-			if(db->config->per_listener_settings && !context->listener && chunk.F.listener_port > 0){
-				for(i=0; i < db->config->listener_count; i++){
-					if(db->config->listeners[i].port == chunk.F.listener_port){
-						context->listener = &db->config->listeners[i];
-						break;
-					}
+		}
+		/* in per_listener_settings mode, try to find the listener by persisted port */
+		if(db.config->per_listener_settings && !context->listener && chunk.F.listener_port > 0){
+			for(i=0; i < db.config->listener_count; i++){
+				if(db.config->listeners[i].port == chunk.F.listener_port){
+					context->listener = &db.config->listeners[i];
+					break;
 				}
 			}
 		}
@@ -222,7 +225,7 @@ static int persist__client_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
 }
 
 
-static int persist__client_msg_chunk_restore(struct mosquitto_db *db, FILE *db_fptr, uint32_t length)
+static int persist__client_msg_chunk_restore(FILE *db_fptr, uint32_t length)
 {
 	struct P_client_msg chunk;
 	int rc;
@@ -235,18 +238,17 @@ static int persist__client_msg_chunk_restore(struct mosquitto_db *db, FILE *db_f
 		rc = persist__chunk_client_msg_read_v234(db_fptr, &chunk);
 	}
 	if(rc){
-		fclose(db_fptr);
 		return rc;
 	}
 
-	rc = persist__client_msg_restore(db, &chunk);
+	rc = persist__client_msg_restore(&chunk);
 	mosquitto__free(chunk.client_id);
 
 	return rc;
 }
 
 
-static int persist__msg_store_chunk_restore(struct mosquitto_db *db, FILE *db_fptr, uint32_t length)
+static int persist__msg_store_chunk_restore(FILE *db_fptr, uint32_t length)
 {
 	struct P_msg_store chunk;
 	struct mosquitto_msg_store *stored = NULL;
@@ -264,25 +266,23 @@ static int persist__msg_store_chunk_restore(struct mosquitto_db *db, FILE *db_fp
 		rc = persist__chunk_msg_store_read_v234(db_fptr, &chunk, db_version);
 	}
 	if(rc){
-		fclose(db_fptr);
 		return rc;
 	}
 
 	if(chunk.F.source_port){
-		for(i=0; i<db->config->listener_count; i++){
-			if(db->config->listeners[i].port == chunk.F.source_port){
-				chunk.source.listener = &db->config->listeners[i];
+		for(i=0; i<db.config->listener_count; i++){
+			if(db.config->listeners[i].port == chunk.F.source_port){
+				chunk.source.listener = &db.config->listeners[i];
 				break;
 			}
 		}
 	}
 	load = mosquitto__calloc(1, sizeof(struct mosquitto_msg_store_load));
 	if(!load){
-		fclose(db_fptr);
 		mosquitto__free(chunk.source.id);
 		mosquitto__free(chunk.source.username);
 		mosquitto__free(chunk.topic);
-		UHPA_FREE(chunk.payload, chunk.F.payloadlen);
+		mosquitto__free(chunk.payload);
 		log__printf(NULL, MOSQ_LOG_ERR, "Error: Out of memory.");
 		return MOSQ_ERR_NOMEM;
 	}
@@ -294,7 +294,7 @@ static int persist__msg_store_chunk_restore(struct mosquitto_db *db, FILE *db_fp
 			mosquitto__free(chunk.source.id);
 			mosquitto__free(chunk.source.username);
 			mosquitto__free(chunk.topic);
-			UHPA_FREE(chunk.payload, chunk.F.payloadlen);
+			mosquitto__free(chunk.payload);
 			mosquitto__free(load);
 			return MOSQ_ERR_SUCCESS;
 		}else{
@@ -304,10 +304,27 @@ static int persist__msg_store_chunk_restore(struct mosquitto_db *db, FILE *db_fp
 		message_expiry_interval = 0;
 	}
 
-	rc = db__message_store(db, &chunk.source, chunk.F.source_mid,
-			chunk.topic, chunk.F.qos, chunk.F.payloadlen,
-			&chunk.payload, chunk.F.retain, &stored, message_expiry_interval,
-			chunk.properties, chunk.F.store_id, mosq_mo_client);
+	stored = mosquitto__calloc(1, sizeof(struct mosquitto_msg_store));
+	if(stored == NULL){
+		mosquitto__free(load);
+		mosquitto__free(chunk.source.id);
+		mosquitto__free(chunk.source.username);
+		mosquitto__free(chunk.topic);
+		mosquitto__free(chunk.payload);
+		log__printf(NULL, MOSQ_LOG_ERR, "Error: Out of memory.");
+		return MOSQ_ERR_NOMEM;
+	}
+
+	stored->source_mid = chunk.F.source_mid;
+	stored->topic = chunk.topic;
+	stored->qos = chunk.F.qos;
+	stored->payloadlen = chunk.F.payloadlen;
+	stored->retain = chunk.F.retain;
+	stored->properties = chunk.properties;
+	stored->payload = chunk.payload;
+
+	rc = db__message_store(&chunk.source, stored, message_expiry_interval,
+			chunk.F.store_id, mosq_mo_client);
 
 	mosquitto__free(chunk.source.id);
 	mosquitto__free(chunk.source.username);
@@ -319,20 +336,21 @@ static int persist__msg_store_chunk_restore(struct mosquitto_db *db, FILE *db_fp
 		load->db_id = stored->db_id;
 		load->store = stored;
 
-		HASH_ADD(hh, db->msg_store_load, db_id, sizeof(dbid_t), load);
+		HASH_ADD(hh, db.msg_store_load, db_id, sizeof(dbid_t), load);
 		return MOSQ_ERR_SUCCESS;
 	}else{
 		mosquitto__free(load);
-		fclose(db_fptr);
 		return rc;
 	}
 }
 
-static int persist__retain_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
+static int persist__retain_chunk_restore(FILE *db_fptr)
 {
 	struct mosquitto_msg_store_load *load;
 	struct P_retain chunk;
 	int rc;
+	char **split_topics;
+	char *local_topic;
 
 	memset(&chunk, 0, sizeof(struct P_retain));
 
@@ -342,20 +360,22 @@ static int persist__retain_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
 		rc = persist__chunk_retain_read_v234(db_fptr, &chunk);
 	}
 	if(rc){
-		fclose(db_fptr);
 		return rc;
 	}
 
-	HASH_FIND(hh, db->msg_store_load, &chunk.F.store_id, sizeof(dbid_t), load);
+	HASH_FIND(hh, db.msg_store_load, &chunk.F.store_id, sizeof(dbid_t), load);
 	if(load){
-		sub__messages_queue(db, NULL, load->store->topic, load->store->qos, load->store->retain, &load->store);
+		if(sub__topic_tokenise(load->store->topic, &local_topic, &split_topics, NULL)) return 1;
+		retain__store(load->store->topic, load->store, split_topics);
+		mosquitto__free(local_topic);
+		mosquitto__free(split_topics);
 	}else{
 		/* Can't find the message - probably expired */
 	}
 	return MOSQ_ERR_SUCCESS;
 }
 
-static int persist__sub_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
+static int persist__sub_chunk_restore(FILE *db_fptr)
 {
 	struct P_sub chunk;
 	int rc;
@@ -368,11 +388,10 @@ static int persist__sub_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
 		rc = persist__chunk_sub_read_v234(db_fptr, &chunk);
 	}
 	if(rc){
-		fclose(db_fptr);
 		return rc;
 	}
 
-	rc = persist__restore_sub(db, chunk.client_id, chunk.topic, chunk.F.qos, chunk.F.identifier, chunk.F.options);
+	rc = persist__restore_sub(chunk.client_id, chunk.topic, chunk.F.qos, chunk.F.identifier, chunk.F.options);
 
 	mosquitto__free(chunk.client_id);
 	mosquitto__free(chunk.topic);
@@ -381,7 +400,7 @@ static int persist__sub_chunk_restore(struct mosquitto_db *db, FILE *db_fptr)
 }
 
 
-int persist__chunk_header_read(FILE *db_fptr, int *chunk, int *length)
+int persist__chunk_header_read(FILE *db_fptr, uint32_t *chunk, uint32_t *length)
 {
 	if(db_version == 6 || db_version == 5){
 		return persist__chunk_header_read_v56(db_fptr, chunk, length);
@@ -391,29 +410,28 @@ int persist__chunk_header_read(FILE *db_fptr, int *chunk, int *length)
 }
 
 
-int persist__restore(struct mosquitto_db *db)
+int persist__restore(void)
 {
 	FILE *fptr;
 	char header[15];
 	int rc = 0;
 	uint32_t crc;
 	uint32_t i32temp;
-	int chunk, length;
-	ssize_t rlen;
+	uint32_t chunk, length;
+	size_t rlen;
 	char *err;
 	struct mosquitto_msg_store_load *load, *load_tmp;
 	struct PF_cfg cfg_chunk;
 
-	assert(db);
-	assert(db->config);
+	assert(db.config);
 
-	if(!db->config->persistence || db->config->persistence_filepath == NULL){
+	if(!db.config->persistence || db.config->persistence_filepath == NULL){
 		return MOSQ_ERR_SUCCESS;
 	}
 
-	db->msg_store_load = NULL;
+	db.msg_store_load = NULL;
 
-	fptr = mosquitto__fopen(db->config->persistence_filepath, "rb", false);
+	fptr = mosquitto__fopen(db.config->persistence_filepath, "rb", false);
 	if(fptr == NULL) return MOSQ_ERR_SUCCESS;
 	rlen = fread(&header, 1, 15, fptr);
 	if(rlen == 0){
@@ -466,27 +484,42 @@ int persist__restore(struct mosquitto_db *db)
 						fclose(fptr);
 						return 1;
 					}
-					db->last_db_id = cfg_chunk.last_db_id;
+					db.last_db_id = cfg_chunk.last_db_id;
 					break;
 
 				case DB_CHUNK_MSG_STORE:
-					if(persist__msg_store_chunk_restore(db, fptr, length)) return 1;
+					if(persist__msg_store_chunk_restore(fptr, length)){
+						fclose(fptr);
+						return 1;
+					}
 					break;
 
 				case DB_CHUNK_CLIENT_MSG:
-					if(persist__client_msg_chunk_restore(db, fptr, length)) return 1;
+					if(persist__client_msg_chunk_restore(fptr, length)){
+						fclose(fptr);
+						return 1;
+					}
 					break;
 
 				case DB_CHUNK_RETAIN:
-					if(persist__retain_chunk_restore(db, fptr)) return 1;
+					if(persist__retain_chunk_restore(fptr)){
+						fclose(fptr);
+						return 1;
+					}
 					break;
 
 				case DB_CHUNK_SUB:
-					if(persist__sub_chunk_restore(db, fptr)) return 1;
+					if(persist__sub_chunk_restore(fptr)){
+						fclose(fptr);
+						return 1;
+					}
 					break;
 
 				case DB_CHUNK_CLIENT:
-					if(persist__client_chunk_restore(db, fptr)) return 1;
+					if(persist__client_chunk_restore(fptr)){
+						fclose(fptr);
+						return 1;
+					}
 					break;
 
 				default:
@@ -495,7 +528,6 @@ int persist__restore(struct mosquitto_db *db)
 					break;
 			}
 		}
-		if(rlen < 0) goto error;
 	}else{
 		log__printf(NULL, MOSQ_LOG_ERR, "Error: Unable to restore persistent database. Unrecognised file format.");
 		rc = 1;
@@ -503,8 +535,8 @@ int persist__restore(struct mosquitto_db *db)
 
 	fclose(fptr);
 
-	HASH_ITER(hh, db->msg_store_load, load, load_tmp){
-		HASH_DELETE(hh, db->msg_store_load, load);
+	HASH_ITER(hh, db.msg_store_load, load, load_tmp){
+		HASH_DELETE(hh, db.msg_store_load, load);
 		mosquitto__free(load);
 	}
 	return rc;
@@ -515,17 +547,16 @@ error:
 	return 1;
 }
 
-static int persist__restore_sub(struct mosquitto_db *db, const char *client_id, const char *sub, int qos, uint32_t identifier, int options)
+static int persist__restore_sub(const char *client_id, const char *sub, uint8_t qos, uint32_t identifier, int options)
 {
 	struct mosquitto *context;
 
-	assert(db);
 	assert(client_id);
 	assert(sub);
 
-	context = persist__find_or_add_context(db, client_id, 0);
+	context = persist__find_or_add_context(client_id, 0);
 	if(!context) return 1;
-	return sub__add(db, context, sub, qos, identifier, options, &db->subs);
+	return sub__add(context, sub, qos, identifier, options, &db.subs);
 }
 
 #endif
