@@ -2,14 +2,16 @@
 Copyright (c) 2011-2020 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
-are made available under the terms of the Eclipse Public License v1.0
+are made available under the terms of the Eclipse Public License 2.0
 and Eclipse Distribution License v1.0 which accompany this distribution.
  
 The Eclipse Public License is available at
-   http://www.eclipse.org/legal/epl-v10.html
+   https://www.eclipse.org/legal/epl-2.0/
 and the Eclipse Distribution License is available at
   http://www.eclipse.org/org/documents/edl-v10.php.
  
+SPDX-License-Identifier: EPL-2.0 OR EDL-1.0
+
 Contributors:
    Roger Light - initial implementation and documentation.
 */
@@ -19,12 +21,15 @@ Contributors:
 #include <stdio.h>
 #include <string.h>
 
+#include "mosquitto_broker.h"
 #include "mosquitto_broker_internal.h"
 #include "mosquitto_plugin.h"
 #include "memory_mosq.h"
 #include "lib_load.h"
+#include "utlist.h"
 
 typedef int (*FUNC_auth_plugin_version)(void);
+typedef int (*FUNC_plugin_version)(int, const int *);
 
 static int security__cleanup_single(struct mosquitto__security_options *opts, bool reload);
 
@@ -51,14 +56,14 @@ int security__load_v2(struct mosquitto__auth_plugin *plugin, struct mosquitto_au
 				"Error: Unable to load auth plugin function mosquitto_auth_plugin_init().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 	if(!(plugin->plugin_cleanup_v2 = (FUNC_auth_plugin_cleanup_v2)LIB_SYM(lib, "mosquitto_auth_plugin_cleanup"))){
 		log__printf(NULL, MOSQ_LOG_ERR,
 				"Error: Unable to load auth plugin function mosquitto_auth_plugin_cleanup().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->security_init_v2 = (FUNC_auth_plugin_security_init_v2)LIB_SYM(lib, "mosquitto_auth_security_init"))){
@@ -66,7 +71,7 @@ int security__load_v2(struct mosquitto__auth_plugin *plugin, struct mosquitto_au
 				"Error: Unable to load auth plugin function mosquitto_auth_security_init().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->security_cleanup_v2 = (FUNC_auth_plugin_security_cleanup_v2)LIB_SYM(lib, "mosquitto_auth_security_cleanup"))){
@@ -74,7 +79,7 @@ int security__load_v2(struct mosquitto__auth_plugin *plugin, struct mosquitto_au
 				"Error: Unable to load auth plugin function mosquitto_auth_security_cleanup().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->acl_check_v2 = (FUNC_auth_plugin_acl_check_v2)LIB_SYM(lib, "mosquitto_auth_acl_check"))){
@@ -82,7 +87,7 @@ int security__load_v2(struct mosquitto__auth_plugin *plugin, struct mosquitto_au
 				"Error: Unable to load auth plugin function mosquitto_auth_acl_check().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->unpwd_check_v2 = (FUNC_auth_plugin_unpwd_check_v2)LIB_SYM(lib, "mosquitto_auth_unpwd_check"))){
@@ -90,7 +95,7 @@ int security__load_v2(struct mosquitto__auth_plugin *plugin, struct mosquitto_au
 				"Error: Unable to load auth plugin function mosquitto_auth_unpwd_check().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->psk_key_get_v2 = (FUNC_auth_plugin_psk_key_get_v2)LIB_SYM(lib, "mosquitto_auth_psk_key_get"))){
@@ -98,7 +103,7 @@ int security__load_v2(struct mosquitto__auth_plugin *plugin, struct mosquitto_au
 				"Error: Unable to load auth plugin function mosquitto_auth_psk_key_get().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	plugin->lib = lib;
@@ -125,14 +130,14 @@ int security__load_v3(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_plugin_init().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 	if(!(plugin->plugin_cleanup_v3 = (FUNC_auth_plugin_cleanup_v3)LIB_SYM(lib, "mosquitto_auth_plugin_cleanup"))){
 		log__printf(NULL, MOSQ_LOG_ERR,
 				"Error: Unable to load auth plugin function mosquitto_auth_plugin_cleanup().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->security_init_v3 = (FUNC_auth_plugin_security_init_v3)LIB_SYM(lib, "mosquitto_auth_security_init"))){
@@ -140,7 +145,7 @@ int security__load_v3(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_security_init().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->security_cleanup_v3 = (FUNC_auth_plugin_security_cleanup_v3)LIB_SYM(lib, "mosquitto_auth_security_cleanup"))){
@@ -148,7 +153,7 @@ int security__load_v3(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_security_cleanup().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->acl_check_v3 = (FUNC_auth_plugin_acl_check_v3)LIB_SYM(lib, "mosquitto_auth_acl_check"))){
@@ -156,7 +161,7 @@ int security__load_v3(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_acl_check().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->unpwd_check_v3 = (FUNC_auth_plugin_unpwd_check_v3)LIB_SYM(lib, "mosquitto_auth_unpwd_check"))){
@@ -164,7 +169,7 @@ int security__load_v3(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_unpwd_check().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->psk_key_get_v3 = (FUNC_auth_plugin_psk_key_get_v3)LIB_SYM(lib, "mosquitto_auth_psk_key_get"))){
@@ -172,7 +177,7 @@ int security__load_v3(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_psk_key_get().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	plugin->lib = lib;
@@ -198,14 +203,14 @@ int security__load_v4(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_plugin_init().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 	if(!(plugin->plugin_cleanup_v4 = (FUNC_auth_plugin_cleanup_v4)LIB_SYM(lib, "mosquitto_auth_plugin_cleanup"))){
 		log__printf(NULL, MOSQ_LOG_ERR,
 				"Error: Unable to load auth plugin function mosquitto_auth_plugin_cleanup().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->security_init_v4 = (FUNC_auth_plugin_security_init_v4)LIB_SYM(lib, "mosquitto_auth_security_init"))){
@@ -213,7 +218,7 @@ int security__load_v4(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_security_init().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->security_cleanup_v4 = (FUNC_auth_plugin_security_cleanup_v4)LIB_SYM(lib, "mosquitto_auth_security_cleanup"))){
@@ -221,7 +226,7 @@ int security__load_v4(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_security_cleanup().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	if(!(plugin->acl_check_v4 = (FUNC_auth_plugin_acl_check_v4)LIB_SYM(lib, "mosquitto_auth_acl_check"))){
@@ -229,7 +234,7 @@ int security__load_v4(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 				"Error: Unable to load auth plugin function mosquitto_auth_acl_check().");
 		LIB_ERROR();
 		LIB_CLOSE(lib);
-		return 1;
+		return MOSQ_ERR_UNKNOWN;
 	}
 
 	plugin->unpwd_check_v4 = (FUNC_auth_plugin_unpwd_check_v4)LIB_SYM(lib, "mosquitto_auth_unpwd_check");
@@ -261,7 +266,7 @@ int security__load_v4(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 			log__printf(NULL, MOSQ_LOG_ERR,
 					"Error: Plugin has missing mosquitto_auth_continue() function.");
 			LIB_CLOSE(lib);
-			return 1;
+			return MOSQ_ERR_UNKNOWN;
 		}
 	}else{
 		log__printf(NULL, MOSQ_LOG_INFO,
@@ -282,13 +287,16 @@ int security__load_v4(struct mosquitto__auth_plugin *plugin, struct mosquitto_op
 }
 
 
-static int security__module_init_single(struct mosquitto__security_options *opts)
+static int security__module_init_single(struct mosquitto__listener *listener, struct mosquitto__security_options *opts)
 {
 	void *lib;
-	int (*plugin_version)(void) = NULL;
+	int (*plugin_version)(int, const int*) = NULL;
+	int (*plugin_auth_version)(void) = NULL;
 	int version;
 	int i;
 	int rc;
+	const int plugin_versions[] = {5, 4, 3, 2};
+	int plugin_version_count = sizeof(plugin_versions)/sizeof(int);
 
 	if(opts->auth_plugin_config_count == 0){
 		return MOSQ_ERR_SUCCESS;
@@ -305,20 +313,34 @@ static int security__module_init_single(struct mosquitto__security_options *opts
 				log__printf(NULL, MOSQ_LOG_ERR,
 						"Error: Unable to load auth plugin \"%s\".", opts->auth_plugin_configs[i].path);
 				LIB_ERROR();
-				return 1;
+				return MOSQ_ERR_UNKNOWN;
 			}
 
 			opts->auth_plugin_configs[i].plugin.lib = NULL;
-			if(!(plugin_version = (FUNC_auth_plugin_version)LIB_SYM(lib, "mosquitto_auth_plugin_version"))){
+			if((plugin_version = (FUNC_plugin_version)LIB_SYM(lib, "mosquitto_plugin_version"))){
+				version = plugin_version(plugin_version_count, plugin_versions);
+			}else if((plugin_auth_version = (FUNC_auth_plugin_version)LIB_SYM(lib, "mosquitto_auth_plugin_version"))){
+				version = plugin_auth_version();
+			}else{
 				log__printf(NULL, MOSQ_LOG_ERR,
-						"Error: Unable to load auth plugin function mosquitto_auth_plugin_version().");
+						"Error: Unable to load auth plugin function mosquitto_auth_plugin_version() or mosquitto_plugin_version().");
 				LIB_ERROR();
 				LIB_CLOSE(lib);
-				return 1;
+				return MOSQ_ERR_UNKNOWN;
 			}
-			version = plugin_version();
 			opts->auth_plugin_configs[i].plugin.version = version;
-			if(version == 4){
+			if(version == 5){
+				rc = plugin__load_v5(
+						listener,
+						&opts->auth_plugin_configs[i].plugin,
+						opts->auth_plugin_configs[i].options,
+						opts->auth_plugin_configs[i].option_count,
+						lib);
+
+				if(rc){
+					return rc;
+				}
+			}else if(version == 4){
 				rc = security__load_v4(
 						&opts->auth_plugin_configs[i].plugin,
 						opts->auth_plugin_configs[i].options,
@@ -350,12 +372,12 @@ static int security__module_init_single(struct mosquitto__security_options *opts
 				}
 			}else{
 				log__printf(NULL, MOSQ_LOG_ERR,
-						"Error: Incorrect auth plugin version (got %d, expected %d).",
-						version, MOSQ_AUTH_PLUGIN_VERSION);
+						"Error: Unsupported auth plugin version (got %d, expected %d).",
+						version, MOSQ_PLUGIN_VERSION);
 				LIB_ERROR();
 
 				LIB_CLOSE(lib);
-				return 1;
+				return MOSQ_ERR_UNKNOWN;
 			}
 		}
 	}
@@ -363,18 +385,18 @@ static int security__module_init_single(struct mosquitto__security_options *opts
 }
 
 
-int mosquitto_security_module_init(struct mosquitto_db *db)
+int mosquitto_security_module_init(void)
 {
 	int rc = MOSQ_ERR_SUCCESS;
 	int i;
 
-	if(db->config->per_listener_settings){
-		for(i=0; i<db->config->listener_count; i++){
-			rc = security__module_init_single(&db->config->listeners[i].security_options);
+	if(db.config->per_listener_settings){
+		for(i=0; i<db.config->listener_count; i++){
+			rc = security__module_init_single(&db.config->listeners[i], &db.config->listeners[i].security_options);
 			if(rc) return rc;
 		}
 	}else{
-		rc = security__module_init_single(&db->config->security_options);
+		rc = security__module_init_single(NULL, &db.config->security_options);
 	}
 	return rc;
 }
@@ -386,7 +408,15 @@ static void security__module_cleanup_single(struct mosquitto__security_options *
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
 		/* Run plugin cleanup function */
-		if(opts->auth_plugin_configs[i].plugin.version == 4){
+		if(opts->auth_plugin_configs[i].plugin.version == 5){
+			opts->auth_plugin_configs[i].plugin.plugin_cleanup_v5(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					opts->auth_plugin_configs[i].options,
+					opts->auth_plugin_configs[i].option_count);
+			mosquitto__free(opts->auth_plugin_configs[i].plugin.identifier);
+			opts->auth_plugin_configs[i].plugin.identifier = NULL;
+
+		}else if(opts->auth_plugin_configs[i].plugin.version == 4){
 			opts->auth_plugin_configs[i].plugin.plugin_cleanup_v4(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					opts->auth_plugin_configs[i].options,
@@ -413,16 +443,16 @@ static void security__module_cleanup_single(struct mosquitto__security_options *
 }
 
 
-int mosquitto_security_module_cleanup(struct mosquitto_db *db)
+int mosquitto_security_module_cleanup(void)
 {
 	int i;
 
-	mosquitto_security_cleanup(db, false);
+	mosquitto_security_cleanup(false);
 
-	security__module_cleanup_single(&db->config->security_options);
+	security__module_cleanup_single(&db.config->security_options);
 
-	for(i=0; i<db->config->listener_count; i++){
-		security__module_cleanup_single(&db->config->listeners[i].security_options);
+	for(i=0; i<db.config->listener_count; i++){
+		security__module_cleanup_single(&db.config->listeners[i].security_options);
 	}
 
 	return MOSQ_ERR_SUCCESS;
@@ -433,9 +463,26 @@ static int security__init_single(struct mosquitto__security_options *opts, bool
 {
 	int i;
 	int rc;
+	struct mosquitto_evt_reload event_data;
+	struct mosquitto__callback *cb_base;
+
+	if(reload){
+		DL_FOREACH(opts->plugin_callbacks.reload, cb_base){
+			memset(&event_data, 0, sizeof(event_data));
+
+			event_data.options = NULL;
+			event_data.option_count = 0;
+			rc = cb_base->cb(MOSQ_EVT_RELOAD, &event_data, cb_base->userdata);
+			if(rc != MOSQ_ERR_PLUGIN_DEFER){
+				return rc;
+			}
+		}
+	}
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
-		if(opts->auth_plugin_configs[i].plugin.version == 4){
+		if(opts->auth_plugin_configs[i].plugin.version == 5){
+			continue;
+		}else if(opts->auth_plugin_configs[i].plugin.version == 4){
 			rc = opts->auth_plugin_configs[i].plugin.security_init_v4(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					opts->auth_plugin_configs[i].options,
@@ -466,21 +513,21 @@ static int security__init_single(struct mosquitto__security_options *opts, bool
 }
 
 
-int mosquitto_security_init(struct mosquitto_db *db, bool reload)
+int mosquitto_security_init(bool reload)
 {
 	int i;
 	int rc;
 
-	if(db->config->per_listener_settings){
-		for(i=0; i<db->config->listener_count; i++){
-			rc = security__init_single(&db->config->listeners[i].security_options, reload);
+	if(db.config->per_listener_settings){
+		for(i=0; i<db.config->listener_count; i++){
+			rc = security__init_single(&db.config->listeners[i].security_options, reload);
 			if(rc != MOSQ_ERR_SUCCESS) return rc;
 		}
 	}else{
-		rc = security__init_single(&db->config->security_options, reload);
+		rc = security__init_single(&db.config->security_options, reload);
 		if(rc != MOSQ_ERR_SUCCESS) return rc;
 	}
-	return mosquitto_security_init_default(db, reload);
+	return mosquitto_security_init_default(reload);
 }
 
 /* Apply security settings after a reload.
@@ -489,9 +536,9 @@ int mosquitto_security_init(struct mosquitto_db *db, bool reload)
  * - Disconnecting users with invalid passwords
  * - Reapplying ACLs
  */
-int mosquitto_security_apply(struct mosquitto_db *db)
+int mosquitto_security_apply(void)
 {
-	return mosquitto_security_apply_default(db);
+	return mosquitto_security_apply_default();
 }
 
 
@@ -501,7 +548,9 @@ static int security__cleanup_single(struct mosquitto__security_options *opts, bo
 	int rc;
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
-		if(opts->auth_plugin_configs[i].plugin.version == 4){
+		if(opts->auth_plugin_configs[i].plugin.version == 5){
+			rc = MOSQ_ERR_SUCCESS;
+		}else if(opts->auth_plugin_configs[i].plugin.version == 4){
 			rc = opts->auth_plugin_configs[i].plugin.security_cleanup_v4(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					opts->auth_plugin_configs[i].options,
@@ -533,23 +582,23 @@ static int security__cleanup_single(struct mosquitto__security_options *opts, bo
 }
 
 
-int mosquitto_security_cleanup(struct mosquitto_db *db, bool reload)
+int mosquitto_security_cleanup(bool reload)
 {
 	int i;
 	int rc;
 
-	rc = security__cleanup_single(&db->config->security_options, reload);
+	rc = security__cleanup_single(&db.config->security_options, reload);
 	if(rc != MOSQ_ERR_SUCCESS) return rc;
 
-	for(i=0; i<db->config->listener_count; i++){
-		rc = security__cleanup_single(&db->config->listeners[i].security_options, reload);
+	for(i=0; i<db.config->listener_count; i++){
+		rc = security__cleanup_single(&db.config->listeners[i].security_options, reload);
 		if(rc != MOSQ_ERR_SUCCESS) return rc;
 	}
-	return mosquitto_security_cleanup_default(db, reload);
+	return mosquitto_security_cleanup_default(reload);
 }
 
 
-/* int mosquitto_acl_check(struct mosquitto_db *db, struct mosquitto *context, const char *topic, int access) */
+/* int mosquitto_acl_check(struct mosquitto *context, const char *topic, int access) */
 static int acl__check_single(struct mosquitto__auth_plugin_config *auth_plugin, struct mosquitto *context, struct mosquitto_acl_msg *msg, int access)
 {
 	const char *username;
@@ -574,11 +623,17 @@ static int acl__check_single(struct mosquitto__auth_plugin_config *auth_plugin,
 	}
 
 	if(auth_plugin->plugin.version == 4){
+		if(access == MOSQ_ACL_UNSUBSCRIBE){
+			return MOSQ_ERR_SUCCESS;
+		}
 		return auth_plugin->plugin.acl_check_v4(auth_plugin->plugin.user_data, access, context, msg);
 	}else if(auth_plugin->plugin.version == 3){
+		if(access == MOSQ_ACL_UNSUBSCRIBE){
+			return MOSQ_ERR_SUCCESS;
+		}
 		return auth_plugin->plugin.acl_check_v3(auth_plugin->plugin.user_data, access, context, msg);
 	}else if(auth_plugin->plugin.version == 2){
-		if(access == MOSQ_ACL_SUBSCRIBE){
+		if(access == MOSQ_ACL_SUBSCRIBE || access == MOSQ_ACL_UNSUBSCRIBE){
 			return MOSQ_ERR_SUCCESS;
 		}
 		return auth_plugin->plugin.acl_check_v2(auth_plugin->plugin.user_data, context->id, username, topic, access);
@@ -609,8 +664,7 @@ static int acl__check_dollar(const char *topic, int access)
 		}
 	}else if(!strncmp(topic, "$share", 6)){
 		/* Only allow sub/unsub to shared subscriptions */
-		if(access == MOSQ_ACL_SUBSCRIBE){
-		/* FIXME if(access == MOSQ_ACL_SUBSCRIBE || access == MOSQ_ACL_UNSUBSCRIBE){ */
+		if(access == MOSQ_ACL_SUBSCRIBE || access == MOSQ_ACL_UNSUBSCRIBE){
 			return MOSQ_ERR_SUCCESS;
 		}else{
 			return MOSQ_ERR_ACL_DENIED;
@@ -622,33 +676,38 @@ static int acl__check_dollar(const char *topic, int access)
 }
 
 
-int mosquitto_acl_check(struct mosquitto_db *db, struct mosquitto *context, const char *topic, long payloadlen, void* payload, int qos, bool retain, int access)
+int mosquitto_acl_check(struct mosquitto *context, const char *topic, uint32_t payloadlen, void* payload, uint8_t qos, bool retain, int access)
 {
 	int rc;
 	int i;
 	struct mosquitto__security_options *opts;
 	struct mosquitto_acl_msg msg;
+	struct mosquitto__callback *cb_base;
+	struct mosquitto_evt_acl_check event_data;
 
 	if(!context->id){
 		return MOSQ_ERR_ACL_DENIED;
 	}
+	if(context->bridge){
+		return MOSQ_ERR_SUCCESS;
+	}
 
 	rc = acl__check_dollar(topic, access);
 	if(rc) return rc;
 
-	rc = mosquitto_acl_check_default(db, context, topic, access);
-	if(rc != MOSQ_ERR_PLUGIN_DEFER){
-		return rc;
-	}
-	/* Default check has accepted or deferred at this point.
+	/* 
 	 * If no plugins exist we should accept at this point so set rc to success.
 	 */
 	rc = MOSQ_ERR_SUCCESS;
 
-	if(db->config->per_listener_settings){
-		opts = &context->listener->security_options;
+	if(db.config->per_listener_settings){
+		if(context->listener){
+			opts = &context->listener->security_options;
+		}else{
+			return MOSQ_ERR_ACL_DENIED;
+		}
 	}else{
-		opts = &db->config->security_options;
+		opts = &db.config->security_options;
 	}
 
 	memset(&msg, 0, sizeof(msg));
@@ -658,13 +717,33 @@ int mosquitto_acl_check(struct mosquitto_db *db, struct mosquitto *context, cons
 	msg.qos = qos;
 	msg.retain = retain;
 
-	for(i=0; i<opts->auth_plugin_config_count; i++){
-		rc = acl__check_single(&opts->auth_plugin_configs[i], context, &msg, access);
+	DL_FOREACH(opts->plugin_callbacks.acl_check, cb_base){
+		/* FIXME - username deny special chars */
+
+		memset(&event_data, 0, sizeof(event_data));
+		event_data.client = context;
+		event_data.access = access;
+		event_data.topic = topic;
+		event_data.payloadlen = payloadlen;
+		event_data.payload = payload;
+		event_data.qos = qos;
+		event_data.retain = retain;
+		event_data.properties = NULL;
+		rc = cb_base->cb(MOSQ_EVT_ACL_CHECK, &event_data, cb_base->userdata);
 		if(rc != MOSQ_ERR_PLUGIN_DEFER){
 			return rc;
 		}
 	}
 
+	for(i=0; i<opts->auth_plugin_config_count; i++){
+		if(opts->auth_plugin_configs[i].plugin.version < 5){
+			rc = acl__check_single(&opts->auth_plugin_configs[i], context, &msg, access);
+			if(rc != MOSQ_ERR_PLUGIN_DEFER){
+				return rc;
+			}
+		}
+	}
+
 	/* If all plugins deferred, this is a denial. If rc == MOSQ_ERR_SUCCESS
 	 * here, then no plugins were configured. */
 	if(rc == MOSQ_ERR_PLUGIN_DEFER){
@@ -673,26 +752,35 @@ int mosquitto_acl_check(struct mosquitto_db *db, struct mosquitto *context, cons
 	return rc;
 }
 
-int mosquitto_unpwd_check(struct mosquitto_db *db, struct mosquitto *context, const char *username, const char *password)
+int mosquitto_unpwd_check(struct mosquitto *context)
 {
 	int rc;
 	int i;
 	struct mosquitto__security_options *opts;
+	struct mosquitto_evt_basic_auth event_data;
+	struct mosquitto__callback *cb_base;
+	bool plugin_used = false;
 
-	rc = mosquitto_unpwd_check_default(db, context, username, password);
-	if(rc != MOSQ_ERR_PLUGIN_DEFER){
-		return rc;
-	}
-	/* Default check has accepted or deferred at this point.
-	 * If no plugins exist we should accept at this point so set rc to success.
-	 */
-	if(db->config->per_listener_settings){
+	rc = MOSQ_ERR_PLUGIN_DEFER;
+
+	if(db.config->per_listener_settings){
 		opts = &context->listener->security_options;
 	}else{
-		opts = &db->config->security_options;
+		opts = &db.config->security_options;
+	}
+
+	DL_FOREACH(opts->plugin_callbacks.basic_auth, cb_base){
+		memset(&event_data, 0, sizeof(event_data));
+		event_data.client = context;
+		event_data.username = context->username;
+		event_data.password = context->password;
+		rc = cb_base->cb(MOSQ_EVT_BASIC_AUTH, &event_data, cb_base->userdata);
+		if(rc != MOSQ_ERR_PLUGIN_DEFER){
+			return rc;
+		}
+		plugin_used = true;
 	}
 
-	rc = MOSQ_ERR_SUCCESS;
 	for(i=0; i<opts->auth_plugin_config_count; i++){
 		if(opts->auth_plugin_configs[i].plugin.version == 4 
 				&& opts->auth_plugin_configs[i].plugin.unpwd_check_v4){
@@ -700,43 +788,62 @@ int mosquitto_unpwd_check(struct mosquitto_db *db, struct mosquitto *context, co
 			rc = opts->auth_plugin_configs[i].plugin.unpwd_check_v4(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					context,
-					username,
-					password);
+					context->username,
+					context->password);
+			plugin_used = true;
 
 		}else if(opts->auth_plugin_configs[i].plugin.version == 3){
 			rc = opts->auth_plugin_configs[i].plugin.unpwd_check_v3(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					context,
-					username,
-					password);
+					context->username,
+					context->password);
+			plugin_used = true;
 
 		}else if(opts->auth_plugin_configs[i].plugin.version == 2){
 			rc = opts->auth_plugin_configs[i].plugin.unpwd_check_v2(
 					opts->auth_plugin_configs[i].plugin.user_data,
-					username,
-					password);
-		}else{
-			rc = MOSQ_ERR_INVAL;
-		}
-		if(rc != MOSQ_ERR_PLUGIN_DEFER){
-			return rc;
+					context->username,
+					context->password);
+			plugin_used = true;
 		}
 	}
 	/* If all plugins deferred, this is a denial. If rc == MOSQ_ERR_SUCCESS
-	 * here, then no plugins were configured. */
-	if(rc == MOSQ_ERR_PLUGIN_DEFER){
-		rc = MOSQ_ERR_AUTH;
+	 * here, then no plugins were configured. Unless we have all deferred, and
+	 * anonymous logins are allowed. */
+	if(plugin_used == false){
+		if((db.config->per_listener_settings && context->listener->security_options.allow_anonymous != false)
+				|| (!db.config->per_listener_settings && db.config->security_options.allow_anonymous != false)){
+
+			return MOSQ_ERR_SUCCESS;
+		}else{
+			return MOSQ_ERR_AUTH;
+		}
+	}else{
+		if(rc == MOSQ_ERR_PLUGIN_DEFER){
+			if(context->username == NULL &&
+					((db.config->per_listener_settings && context->listener->security_options.allow_anonymous != false)
+					|| (!db.config->per_listener_settings && db.config->security_options.allow_anonymous != false))){
+	
+				return MOSQ_ERR_SUCCESS;
+			}else{
+				return MOSQ_ERR_AUTH;
+			}
+		}
 	}
+
 	return rc;
 }
 
-int mosquitto_psk_key_get(struct mosquitto_db *db, struct mosquitto *context, const char *hint, const char *identity, char *key, int max_key_len)
+int mosquitto_psk_key_get(struct mosquitto *context, const char *hint, const char *identity, char *key, int max_key_len)
 {
 	int rc;
 	int i;
 	struct mosquitto__security_options *opts;
+	struct mosquitto_evt_psk_key event_data;
+	struct mosquitto__callback *cb_base;
 
-	rc = mosquitto_psk_key_get_default(db, context, hint, identity, key, max_key_len);
+	rc = mosquitto_psk_key_get_default(context, hint, identity, key, max_key_len);
 	if(rc != MOSQ_ERR_PLUGIN_DEFER){
 		return rc;
 	}
@@ -745,10 +852,23 @@ int mosquitto_psk_key_get(struct mosquitto_db *db, struct mosquitto *context, co
 	 * If no plugins exist we should accept at this point so set rc to success.
 	 */
 
-	if(db->config->per_listener_settings){
+	if(db.config->per_listener_settings){
 		opts = &context->listener->security_options;
 	}else{
-		opts = &db->config->security_options;
+		opts = &db.config->security_options;
+	}
+
+	DL_FOREACH(opts->plugin_callbacks.psk_key, cb_base){
+		memset(&event_data, 0, sizeof(event_data));
+		event_data.client = context;
+		event_data.hint = hint;
+		event_data.identity = identity;
+		event_data.key = key;
+		event_data.max_key_len = max_key_len;
+		rc = cb_base->cb(MOSQ_EVT_PSK_KEY, &event_data, cb_base->userdata);
+		if(rc != MOSQ_ERR_PLUGIN_DEFER){
+			return rc;
+		}
 	}
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
@@ -796,19 +916,36 @@ int mosquitto_psk_key_get(struct mosquitto_db *db, struct mosquitto *context, co
 }
 
 
-int mosquitto_security_auth_start(struct mosquitto_db *db, struct mosquitto *context, bool reauth, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len)
+int mosquitto_security_auth_start(struct mosquitto *context, bool reauth, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len)
 {
 	int rc = MOSQ_ERR_PLUGIN_DEFER;
 	int i;
 	struct mosquitto__security_options *opts;
+	struct mosquitto_evt_extended_auth event_data;
+	struct mosquitto__callback *cb_base;
 
 	if(!context || !context->listener || !context->auth_method) return MOSQ_ERR_INVAL;
 	if(!data_out || !data_out_len) return MOSQ_ERR_INVAL;
 
-	if(db->config->per_listener_settings){
+	if(db.config->per_listener_settings){
 		opts = &context->listener->security_options;
 	}else{
-		opts = &db->config->security_options;
+		opts = &db.config->security_options;
+	}
+
+	DL_FOREACH(opts->plugin_callbacks.ext_auth_start, cb_base){
+		memset(&event_data, 0, sizeof(event_data));
+		event_data.client = context;
+		event_data.data_in = data_in;
+		event_data.data_out = NULL;
+		event_data.data_in_len = data_in_len;
+		event_data.data_out_len = 0;
+		rc = cb_base->cb(MOSQ_EVT_EXT_AUTH_START, &event_data, cb_base->userdata);
+		if(rc != MOSQ_ERR_PLUGIN_DEFER){
+			*data_out = event_data.data_out;
+			*data_out_len = event_data.data_out_len;
+			return rc;
+		}
 	}
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
@@ -838,23 +975,40 @@ int mosquitto_security_auth_start(struct mosquitto_db *db, struct mosquitto *con
 }
 
 
-int mosquitto_security_auth_continue(struct mosquitto_db *db, struct mosquitto *context, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len)
+int mosquitto_security_auth_continue(struct mosquitto *context, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len)
 {
 	int rc = MOSQ_ERR_PLUGIN_DEFER;
 	int i;
 	struct mosquitto__security_options *opts;
+	struct mosquitto_evt_extended_auth event_data;
+	struct mosquitto__callback *cb_base;
 
 	if(!context || !context->listener || !context->auth_method) return MOSQ_ERR_INVAL;
 	if(!data_out || !data_out_len) return MOSQ_ERR_INVAL;
 
-	if(db->config->per_listener_settings){
+	if(db.config->per_listener_settings){
 		opts = &context->listener->security_options;
 	}else{
-		opts = &db->config->security_options;
+		opts = &db.config->security_options;
+	}
+
+	DL_FOREACH(opts->plugin_callbacks.ext_auth_continue, cb_base){
+		memset(&event_data, 0, sizeof(event_data));
+		event_data.client = context;
+		event_data.data_in = data_in;
+		event_data.data_out = NULL;
+		event_data.data_in_len = data_in_len;
+		event_data.data_out_len = 0;
+		rc = cb_base->cb(MOSQ_EVT_EXT_AUTH_CONTINUE, &event_data, cb_base->userdata);
+		if(rc != MOSQ_ERR_PLUGIN_DEFER){
+			*data_out = event_data.data_out;
+			*data_out_len = event_data.data_out_len;
+			return rc;
+		}
 	}
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
-		if(opts->auth_plugin_configs[i].plugin.auth_start_v4){
+		if(opts->auth_plugin_configs[i].plugin.auth_continue_v4){
 			*data_out = NULL;
 			*data_out_len = 0;
 
