@@ -2,14 +2,16 @@
 Copyright (c) 2010-2020 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
-are made available under the terms of the Eclipse Public License v1.0
+are made available under the terms of the Eclipse Public License 2.0
 and Eclipse Distribution License v1.0 which accompany this distribution.
  
 The Eclipse Public License is available at
-   http://www.eclipse.org/legal/epl-v10.html
+   https://www.eclipse.org/legal/epl-2.0/
 and the Eclipse Distribution License is available at
   http://www.eclipse.org/org/documents/edl-v10.php.
  
+SPDX-License-Identifier: EPL-2.0 OR EDL-1.0
+
 Contributors:
    Roger Light - initial implementation and documentation.
 */
@@ -58,29 +60,22 @@ Contributors:
 
 #include "utlist.h"
 
-struct sub__token {
-	struct sub__token *next;
-	char *topic;
-	uint16_t topic_len;
-};
-
-
-static int subs__send(struct mosquitto_db *db, struct mosquitto__subleaf *leaf, const char *topic, int qos, int retain, struct mosquitto_msg_store *stored)
+static int subs__send(struct mosquitto__subleaf *leaf, const char *topic, uint8_t qos, int retain, struct mosquitto_msg_store *stored)
 {
 	bool client_retain;
 	uint16_t mid;
-	int client_qos, msg_qos;
+	uint8_t client_qos, msg_qos;
 	mosquitto_property *properties = NULL;
 	int rc2;
 
 	/* Check for ACL topic access. */
-	rc2 = mosquitto_acl_check(db, leaf->context, topic, stored->payloadlen, UHPA_ACCESS(stored->payload, stored->payloadlen), stored->qos, stored->retain, MOSQ_ACL_READ);
+	rc2 = mosquitto_acl_check(leaf->context, topic, stored->payloadlen, stored->payload, stored->qos, stored->retain, MOSQ_ACL_READ);
 	if(rc2 == MOSQ_ERR_ACL_DENIED){
 		return MOSQ_ERR_SUCCESS;
 	}else if(rc2 == MOSQ_ERR_SUCCESS){
 		client_qos = leaf->qos;
 
-		if(db->config->upgrade_outgoing_qos){
+		if(db.config->upgrade_outgoing_qos){
 			msg_qos = client_qos;
 		}else{
 			if(qos > client_qos){
@@ -102,7 +97,7 @@ static int subs__send(struct mosquitto_db *db, struct mosquitto__subleaf *leaf,
 		if(leaf->identifier){
 			mosquitto_property_add_varint(&properties, MQTT_PROP_SUBSCRIPTION_IDENTIFIER, leaf->identifier);
 		}
-		if(db__message_insert(db, leaf->context, mid, mosq_md_out, msg_qos, client_retain, stored, properties) == 1){
+		if(db__message_insert(leaf->context, mid, mosq_md_out, msg_qos, client_retain, stored, properties, true) == 1){
 			return 1;
 		}
 	}else{
@@ -112,7 +107,7 @@ static int subs__send(struct mosquitto_db *db, struct mosquitto__subleaf *leaf,
 }
 
 
-static int subs__shared_process(struct mosquitto_db *db, struct mosquitto__subhier *hier, const char *topic, int qos, int retain, struct mosquitto_msg_store *stored)
+static int subs__shared_process(struct mosquitto__subhier *hier, const char *topic, uint8_t qos, int retain, struct mosquitto_msg_store *stored)
 {
 	int rc = 0, rc2;
 	struct mosquitto__subshared *shared, *shared_tmp;
@@ -120,7 +115,7 @@ static int subs__shared_process(struct mosquitto_db *db, struct mosquitto__subhi
 
 	HASH_ITER(hh, hier->shared, shared, shared_tmp){
 		leaf = shared->subs;
-		rc2 = subs__send(db, leaf, topic, qos, retain, stored);
+		rc2 = subs__send(leaf, topic, qos, retain, stored);
 		/* Remove current from the top, add back to the bottom */
 		DL_DELETE(shared->subs, leaf);
 		DL_APPEND(shared->subs, leaf);
@@ -131,38 +126,13 @@ static int subs__shared_process(struct mosquitto_db *db, struct mosquitto__subhi
 	return rc;
 }
 
-static int subs__process(struct mosquitto_db *db, struct mosquitto__subhier *hier, const char *source_id, const char *topic, int qos, int retain, struct mosquitto_msg_store *stored, bool set_retain)
+static int subs__process(struct mosquitto__subhier *hier, const char *source_id, const char *topic, uint8_t qos, int retain, struct mosquitto_msg_store *stored)
 {
 	int rc = 0;
 	int rc2;
 	struct mosquitto__subleaf *leaf;
 
-	if(retain && set_retain){
-#ifdef WITH_PERSISTENCE
-		if(strncmp(topic, "$SYS", 4)){
-			/* Retained messages count as a persistence change, but only if
-			 * they aren't for $SYS. */
-			db->persistence_changes++;
-		}
-#endif
-		if(hier->retained){
-			db__msg_store_ref_dec(db, &hier->retained);
-#ifdef WITH_SYS_TREE
-			db->retained_count--;
-#endif
-		}
-		if(stored->payloadlen){
-			hier->retained = stored;
-			db__msg_store_ref_inc(hier->retained);
-#ifdef WITH_SYS_TREE
-			db->retained_count++;
-#endif
-		}else{
-			hier->retained = NULL;
-		}
-	}
-
-	rc = subs__shared_process(db, hier, topic, qos, retain, stored);
+	rc = subs__shared_process(hier, topic, qos, retain, stored);
 
 	leaf = hier->subs;
 	while(source_id && leaf){
@@ -170,7 +140,7 @@ static int subs__process(struct mosquitto_db *db, struct mosquitto__subhier *hie
 			leaf = leaf->next;
 			continue;
 		}
-		rc2 = subs__send(db, leaf, topic, qos, retain, stored);
+		rc2 = subs__send(leaf, topic, qos, retain, stored);
 		if(rc2){
 			rc = 1;
 		}
@@ -183,126 +153,8 @@ static int subs__process(struct mosquitto_db *db, struct mosquitto__subhier *hie
 	}
 }
 
-static struct sub__token *sub__topic_append(struct sub__token **tail, struct sub__token **topics, char *topic)
-{
-	struct sub__token *new_topic;
 
-	if(!topic){
-		return NULL;
-	}
-	new_topic = mosquitto__malloc(sizeof(struct sub__token));
-	if(!new_topic){
-		return NULL;
-	}
-	new_topic->next = NULL;
-	new_topic->topic_len = strlen(topic);
-	new_topic->topic = mosquitto__malloc(new_topic->topic_len+1);
-	if(!new_topic->topic){
-		mosquitto__free(new_topic);
-		return NULL;
-	}
-	strncpy(new_topic->topic, topic, new_topic->topic_len+1);
-
-	if(*tail){
-		(*tail)->next = new_topic;
-		*tail = (*tail)->next;
-	}else{
-		*topics = new_topic;
-		*tail = new_topic;
-	}
-	return new_topic;
-}
-
-static int sub__topic_tokenise(const char *subtopic, struct sub__token **topics)
-{
-	struct sub__token *new_topic, *tail = NULL;
-	size_t len;
-	size_t start, stop, tlen;
-	int i;
-	char *topic;
-	int count = 0;
-
-	assert(subtopic);
-	assert(topics);
-
-	len = strlen(subtopic);
-	if(len == 0){
-		return 1;
-	}
-
-	if(subtopic[0] != '$'){
-		new_topic = sub__topic_append(&tail, topics, "");
-		if(!new_topic) goto cleanup;
-	}
-
-	if(subtopic[0] == '/'){
-		new_topic = sub__topic_append(&tail, topics, "");
-		if(!new_topic) goto cleanup;
-
-		start = 1;
-	}else{
-		start = 0;
-	}
-
-	for(i=start; i<len+1; i++){
-		if(subtopic[i] == '/' || subtopic[i] == '\0'){
-			stop = i;
-			count++;
-
-			if(start != stop){
-				tlen = stop-start;
-
-				topic = mosquitto__malloc(tlen+1);
-				if(!topic) goto cleanup;
-				memcpy(topic, &subtopic[start], tlen);
-				topic[tlen] = '\0';
-				new_topic = sub__topic_append(&tail, topics, topic);
-				mosquitto__free(topic);
-			}else{
-				new_topic = sub__topic_append(&tail, topics, "");
-			}
-			if(!new_topic) goto cleanup;
-			start = i+1;
-		}
-	}
-
-	if(count > TOPIC_HIERARCHY_LIMIT){
-		/* Set limit on hierarchy levels, to restrict stack usage. */
-		goto cleanup;
-	}
-
-	if(*topics != NULL){
-		return MOSQ_ERR_SUCCESS;
-	}else{
-		return 1;
-	}
-
-cleanup:
-	tail = *topics;
-	*topics = NULL;
-	while(tail){
-		mosquitto__free(tail->topic);
-		new_topic = tail->next;
-		mosquitto__free(tail);
-		tail = new_topic;
-	}
-	return 1;
-}
-
-static void sub__topic_tokens_free(struct sub__token *tokens)
-{
-	struct sub__token *tail;
-
-	while(tokens){
-		tail = tokens->next;
-		mosquitto__free(tokens->topic);
-		mosquitto__free(tokens);
-		tokens = tail;
-	}
-}
-
-
-static int sub__add_leaf(struct mosquitto *context, int qos, uint32_t identifier, int options, struct mosquitto__subleaf **head, struct mosquitto__subleaf **newleaf)
+static int sub__add_leaf(struct mosquitto *context, uint8_t qos, uint32_t identifier, int options, struct mosquitto__subleaf **head, struct mosquitto__subleaf **newleaf)
 {
 	struct mosquitto__subleaf *leaf;
 
@@ -347,28 +199,29 @@ static void sub__remove_shared_leaf(struct mosquitto__subhier *subhier, struct m
 }
 
 
-static int sub__add_shared(struct mosquitto_db *db, struct mosquitto *context, int qos, uint32_t identifier, int options, struct mosquitto__subhier *subhier, char *sharename)
+static int sub__add_shared(struct mosquitto *context, uint8_t qos, uint32_t identifier, int options, struct mosquitto__subhier *subhier, const char *sharename)
 {
 	struct mosquitto__subleaf *newleaf;
 	struct mosquitto__subshared *shared = NULL;
 	struct mosquitto__subshared_ref **shared_subs;
 	struct mosquitto__subshared_ref *shared_ref;
 	int i;
-	unsigned int slen;
+	size_t slen;
 	int rc;
 
 	slen = strlen(sharename);
 
 	HASH_FIND(hh, subhier->shared, sharename, slen, shared);
-	if(shared){
-		mosquitto__free(sharename);
-	}else{
+	if(shared == NULL){
 		shared = mosquitto__calloc(1, sizeof(struct mosquitto__subshared));
 		if(!shared){
-			mosquitto__free(sharename);
 			return MOSQ_ERR_NOMEM;
 		}
-		shared->name = sharename;
+		shared->name = mosquitto__strdup(sharename);
+		if(shared->name == NULL){
+			mosquitto__free(shared);
+			return MOSQ_ERR_NOMEM;
+		}
 
 		HASH_ADD_KEYPTR(hh, subhier->shared, shared->name, slen, shared);
 	}
@@ -395,12 +248,15 @@ static int sub__add_shared(struct mosquitto_db *db, struct mosquitto *context, i
 		for(i=0; i<context->shared_sub_count; i++){
 			if(!context->shared_subs[i]){
 				context->shared_subs[i] = shared_ref;
+				shared_ref = NULL;
 				break;
 			}
 		}
-		if(i == context->shared_sub_count){
-			shared_subs = mosquitto__realloc(context->shared_subs, sizeof(struct mosquitto__subhier_ref *)*(context->shared_sub_count + 1));
+		if(shared_ref){
+			shared_subs = mosquitto__realloc(context->shared_subs, sizeof(struct mosquitto__subhier_ref *)*(size_t)(context->shared_sub_count + 1));
 			if(!shared_subs){
+				mosquitto__free(shared_ref);
+				context->shared_subs[context->shared_sub_count-1] = NULL;
 				sub__remove_shared_leaf(subhier, shared, newleaf);
 				return MOSQ_ERR_NOMEM;
 			}
@@ -409,7 +265,7 @@ static int sub__add_shared(struct mosquitto_db *db, struct mosquitto *context, i
 			context->shared_subs[context->shared_sub_count-1] = shared_ref;
 		}
 #ifdef WITH_SYS_TREE
-		db->shared_subscription_count++;
+		db.shared_subscription_count++;
 #endif
 	}
 
@@ -423,7 +279,7 @@ static int sub__add_shared(struct mosquitto_db *db, struct mosquitto *context, i
 }
 
 
-static int sub__add_normal(struct mosquitto_db *db, struct mosquitto *context, int qos, uint32_t identifier, int options, struct mosquitto__subhier *subhier)
+static int sub__add_normal(struct mosquitto *context, uint8_t qos, uint32_t identifier, int options, struct mosquitto__subhier *subhier)
 {
 	struct mosquitto__subleaf *newleaf = NULL;
 	struct mosquitto__subhier **subs;
@@ -443,7 +299,7 @@ static int sub__add_normal(struct mosquitto_db *db, struct mosquitto *context, i
 			}
 		}
 		if(i == context->sub_count){
-			subs = mosquitto__realloc(context->subs, sizeof(struct mosquitto__subhier *)*(context->sub_count + 1));
+			subs = mosquitto__realloc(context->subs, sizeof(struct mosquitto__subhier *)*(size_t)(context->sub_count + 1));
 			if(!subs){
 				DL_DELETE(subhier->subs, newleaf);
 				mosquitto__free(newleaf);
@@ -454,7 +310,7 @@ static int sub__add_normal(struct mosquitto_db *db, struct mosquitto *context, i
 			context->subs[context->sub_count-1] = subhier;
 		}
 #ifdef WITH_SYS_TREE
-		db->subscription_count++;
+		db.subscription_count++;
 #endif
 	}
 
@@ -468,28 +324,34 @@ static int sub__add_normal(struct mosquitto_db *db, struct mosquitto *context, i
 }
 
 
-static int sub__add_context(struct mosquitto_db *db, struct mosquitto *context, int qos, uint32_t identifier, int options, struct mosquitto__subhier *subhier, struct sub__token *tokens, char *sharename)
+static int sub__add_context(struct mosquitto *context, uint8_t qos, uint32_t identifier, int options, struct mosquitto__subhier *subhier, char *const *const topics, const char *sharename)
 {
 	struct mosquitto__subhier *branch;
+	int topic_index = 0;
+	size_t topiclen;
 
 	/* Find leaf node */
-	while(tokens){
-		HASH_FIND(hh, subhier->children, tokens->topic, tokens->topic_len, branch);
+	while(topics && topics[topic_index] != NULL){
+		topiclen = strlen(topics[topic_index]);
+		if(topiclen > UINT16_MAX){
+			return MOSQ_ERR_INVAL;
+		}
+		HASH_FIND(hh, subhier->children, topics[topic_index], topiclen, branch);
 		if(!branch){
 			/* Not found */
-			branch = sub__add_hier_entry(subhier, &subhier->children, tokens->topic, tokens->topic_len);
+			branch = sub__add_hier_entry(subhier, &subhier->children, topics[topic_index], (uint16_t)topiclen);
 			if(!branch) return MOSQ_ERR_NOMEM;
 		}
 		subhier = branch;
-		tokens = tokens ->next;
+		topic_index++;
 	}
 
 	/* Add add our context */
 	if(context && context->id){
 		if(sharename){
-			return sub__add_shared(db, context, qos, identifier, options, subhier, sharename);
+			return sub__add_shared(context, qos, identifier, options, subhier, sharename);
 		}else{
-			return sub__add_normal(db, context, qos, identifier, options, subhier);
+			return sub__add_normal(context, qos, identifier, options, subhier);
 		}
 	}else{
 		return MOSQ_ERR_SUCCESS;
@@ -497,7 +359,7 @@ static int sub__add_context(struct mosquitto_db *db, struct mosquitto *context,
 }
 
 
-static int sub__remove_normal(struct mosquitto_db *db, struct mosquitto *context, struct mosquitto__subhier *subhier, uint8_t *reason)
+static int sub__remove_normal(struct mosquitto *context, struct mosquitto__subhier *subhier, uint8_t *reason)
 {
 	struct mosquitto__subleaf *leaf;
 	int i;
@@ -506,7 +368,7 @@ static int sub__remove_normal(struct mosquitto_db *db, struct mosquitto *context
 	while(leaf){
 		if(leaf->context==context){
 #ifdef WITH_SYS_TREE
-			db->subscription_count--;
+			db.subscription_count--;
 #endif
 			DL_DELETE(subhier->subs, leaf);
 			mosquitto__free(leaf);
@@ -530,20 +392,19 @@ static int sub__remove_normal(struct mosquitto_db *db, struct mosquitto *context
 }
 
 
-static int sub__remove_shared(struct mosquitto_db *db, struct mosquitto *context, struct mosquitto__subhier *subhier, uint8_t *reason, char *sharename)
+static int sub__remove_shared(struct mosquitto *context, struct mosquitto__subhier *subhier, uint8_t *reason, const char *sharename)
 {
 	struct mosquitto__subshared *shared;
 	struct mosquitto__subleaf *leaf;
 	int i;
 
 	HASH_FIND(hh, subhier->shared, sharename, strlen(sharename), shared);
-	mosquitto__free(sharename);
 	if(shared){
 		leaf = shared->subs;
 		while(leaf){
 			if(leaf->context==context){
 #ifdef WITH_SYS_TREE
-				db->shared_subscription_count--;
+				db.shared_subscription_count--;
 #endif
 				DL_DELETE(shared->subs, leaf);
 				mosquitto__free(leaf);
@@ -581,22 +442,22 @@ static int sub__remove_shared(struct mosquitto_db *db, struct mosquitto *context
 }
 
 
-static int sub__remove_recurse(struct mosquitto_db *db, struct mosquitto *context, struct mosquitto__subhier *subhier, struct sub__token *tokens, uint8_t *reason, char *sharename)
+static int sub__remove_recurse(struct mosquitto *context, struct mosquitto__subhier *subhier, char **topics, uint8_t *reason, const char *sharename)
 {
 	struct mosquitto__subhier *branch;
 
-	if(!tokens){
+	if(topics == NULL || topics[0] == NULL){
 		if(sharename){
-			return sub__remove_shared(db, context, subhier, reason, sharename);
+			return sub__remove_shared(context, subhier, reason, sharename);
 		}else{
-			return sub__remove_normal(db, context, subhier, reason);
+			return sub__remove_normal(context, subhier, reason);
 		}
 	}
 
-	HASH_FIND(hh, subhier->children, tokens->topic, tokens->topic_len, branch);
+	HASH_FIND(hh, subhier->children, topics[0], strlen(topics[0]), branch);
 	if(branch){
-		sub__remove_recurse(db, context, branch, tokens->next, reason, sharename);
-		if(!branch->children && !branch->subs && !branch->retained && !branch->shared){
+		sub__remove_recurse(context, branch, &(topics[1]), reason, sharename);
+		if(!branch->children && !branch->subs && !branch->shared){
 			HASH_DELETE(hh, subhier->children, branch);
 			mosquitto__free(branch->topic);
 			mosquitto__free(branch);
@@ -605,26 +466,27 @@ static int sub__remove_recurse(struct mosquitto_db *db, struct mosquitto *contex
 	return MOSQ_ERR_SUCCESS;
 }
 
-static int sub__search(struct mosquitto_db *db, struct mosquitto__subhier *subhier, struct sub__token *tokens, const char *source_id, const char *topic, int qos, int retain, struct mosquitto_msg_store *stored, bool set_retain)
+
+static int sub__search(struct mosquitto__subhier *subhier, char **split_topics, const char *source_id, const char *topic, uint8_t qos, int retain, struct mosquitto_msg_store *stored)
 {
 	/* FIXME - need to take into account source_id if the client is a bridge */
 	struct mosquitto__subhier *branch;
 	int rc;
 	bool have_subscribers = false;
 
-	if(tokens){
+	if(split_topics && split_topics[0]){
 		/* Check for literal match */
-		HASH_FIND(hh, subhier->children, tokens->topic, tokens->topic_len, branch);
+		HASH_FIND(hh, subhier->children, split_topics[0], strlen(split_topics[0]), branch);
 
 		if(branch){
-			rc = sub__search(db, branch, tokens->next, source_id, topic, qos, retain, stored, set_retain);
+			rc = sub__search(branch, &(split_topics[1]), source_id, topic, qos, retain, stored);
 			if(rc == MOSQ_ERR_SUCCESS){
 				have_subscribers = true;
 			}else if(rc != MOSQ_ERR_NO_SUBSCRIBERS){
 				return rc;
 			}
-			if(!tokens->next){
-				rc = subs__process(db, branch, source_id, topic, qos, retain, stored, set_retain);
+			if(split_topics[1] == NULL){ /* End of list */
+				rc = subs__process(branch, source_id, topic, qos, retain, stored);
 				if(rc == MOSQ_ERR_SUCCESS){
 					have_subscribers = true;
 				}else if(rc != MOSQ_ERR_NO_SUBSCRIBERS){
@@ -637,14 +499,14 @@ static int sub__search(struct mosquitto_db *db, struct mosquitto__subhier *subhi
 		HASH_FIND(hh, subhier->children, "+", 1, branch);
 
 		if(branch){
-			rc = sub__search(db, branch, tokens->next, source_id, topic, qos, retain, stored, false);
+			rc = sub__search(branch, &(split_topics[1]), source_id, topic, qos, retain, stored);
 			if(rc == MOSQ_ERR_SUCCESS){
 				have_subscribers = true;
 			}else if(rc != MOSQ_ERR_NO_SUBSCRIBERS){
 				return rc;
 			}
-			if(!tokens->next){
-				rc = subs__process(db, branch, source_id, topic, qos, retain, stored, false);
+			if(split_topics[1] == NULL){ /* End of list */
+				rc = subs__process(branch, source_id, topic, qos, retain, stored);
 				if(rc == MOSQ_ERR_SUCCESS){
 					have_subscribers = true;
 				}else if(rc != MOSQ_ERR_NO_SUBSCRIBERS){
@@ -661,7 +523,7 @@ static int sub__search(struct mosquitto_db *db, struct mosquitto__subhier *subhi
 		 * subscriptions but *don't* return. Although this branch has ended
 		 * there may still be other subscriptions to deal with.
 		 */
-		rc = subs__process(db, branch, source_id, topic, qos, retain, stored, false);
+		rc = subs__process(branch, source_id, topic, qos, retain, stored);
 		if(rc == MOSQ_ERR_SUCCESS){
 			have_subscribers = true;
 		}else if(rc != MOSQ_ERR_NO_SUBSCRIBERS){
@@ -677,7 +539,7 @@ static int sub__search(struct mosquitto_db *db, struct mosquitto__subhier *subhi
 }
 
 
-struct mosquitto__subhier *sub__add_hier_entry(struct mosquitto__subhier *parent, struct mosquitto__subhier **sibling, const char *topic, size_t len)
+struct mosquitto__subhier *sub__add_hier_entry(struct mosquitto__subhier *parent, struct mosquitto__subhier **sibling, const char *topic, uint16_t len)
 {
 	struct mosquitto__subhier *child;
 
@@ -690,14 +552,12 @@ struct mosquitto__subhier *sub__add_hier_entry(struct mosquitto__subhier *parent
 	}
 	child->parent = parent;
 	child->topic_len = len;
-	child->topic = mosquitto__malloc(len+1);
+	child->topic = mosquitto__strdup(topic);
 	if(!child->topic){
 		child->topic_len = 0;
 		mosquitto__free(child);
 		log__printf(NULL, MOSQ_LOG_ERR, "Error: Out of memory.");
 		return NULL;
-	}else{
-		strncpy(child->topic, topic, child->topic_len+1);
 	}
 
 	HASH_ADD_KEYPTR(hh, *sibling, child->topic, child->topic_len, child);
@@ -706,111 +566,83 @@ struct mosquitto__subhier *sub__add_hier_entry(struct mosquitto__subhier *parent
 }
 
 
-int sub__add(struct mosquitto_db *db, struct mosquitto *context, const char *sub, int qos, uint32_t identifier, int options, struct mosquitto__subhier **root)
+int sub__add(struct mosquitto *context, const char *sub, uint8_t qos, uint32_t identifier, int options, struct mosquitto__subhier **root)
 {
 	int rc = 0;
 	struct mosquitto__subhier *subhier;
-	struct sub__token *tokens = NULL, *t;
-	char *sharename = NULL;
+	const char *sharename = NULL;
+	char *local_sub;
+	char **topics;
+	size_t topiclen;
 
 	assert(root);
 	assert(*root);
 	assert(sub);
 
-	if(sub__topic_tokenise(sub, &tokens)) return 1;
+	rc = sub__topic_tokenise(sub, &local_sub, &topics, &sharename);
+	if(rc) return rc;
 
-	if(!strcmp(tokens->topic, "$share")){
-		if(!tokens->next || !tokens->next->next){
-			sub__topic_tokens_free(tokens);
-			return MOSQ_ERR_PROTOCOL;
-		}
-		t = tokens->next;
-		mosquitto__free(tokens->topic);
-		mosquitto__free(tokens);
-		tokens = t;
-
-		sharename = tokens->topic;
-
-		tokens->topic = mosquitto__strdup("");
-		if(!tokens->topic){
-			tokens->topic = sharename;
-			sub__topic_tokens_free(tokens);
-			return MOSQ_ERR_PROTOCOL;
-		}
-		tokens->topic_len = 0;
+	topiclen = strlen(topics[0]);
+	if(topiclen > UINT16_MAX){
+		mosquitto__free(local_sub);
+		mosquitto__free(topics);
+		return MOSQ_ERR_INVAL;
 	}
-
-	HASH_FIND(hh, *root, tokens->topic, tokens->topic_len, subhier);
+	HASH_FIND(hh, *root, topics[0], topiclen, subhier);
 	if(!subhier){
-		subhier = sub__add_hier_entry(NULL, root, tokens->topic, tokens->topic_len);
+		subhier = sub__add_hier_entry(NULL, root, topics[0], (uint16_t)topiclen);
 		if(!subhier){
-			sub__topic_tokens_free(tokens);
+			mosquitto__free(local_sub);
+			mosquitto__free(topics);
 			log__printf(NULL, MOSQ_LOG_ERR, "Error: Out of memory.");
 			return MOSQ_ERR_NOMEM;
 		}
 
 	}
-	rc = sub__add_context(db, context, qos, identifier, options, subhier, tokens, sharename);
+	rc = sub__add_context(context, qos, identifier, options, subhier, topics, sharename);
 
-	sub__topic_tokens_free(tokens);
+	mosquitto__free(local_sub);
+	mosquitto__free(topics);
 
 	return rc;
 }
 
-int sub__remove(struct mosquitto_db *db, struct mosquitto *context, const char *sub, struct mosquitto__subhier *root, uint8_t *reason)
+int sub__remove(struct mosquitto *context, const char *sub, struct mosquitto__subhier *root, uint8_t *reason)
 {
 	int rc = 0;
 	struct mosquitto__subhier *subhier;
-	struct sub__token *tokens = NULL, *t;
-	char *sharename = NULL;
+	const char *sharename = NULL;
+	char *local_sub = NULL;
+	char **topics = NULL;
 
 	assert(root);
 	assert(sub);
 
-	if(sub__topic_tokenise(sub, &tokens)) return 1;
+	rc = sub__topic_tokenise(sub, &local_sub, &topics, &sharename);
+	if(rc) return rc;
 
-	if(!strcmp(tokens->topic, "$share")){
-		if(!tokens->next || !tokens->next->next){
-			sub__topic_tokens_free(tokens);
-			return MOSQ_ERR_PROTOCOL;
-		}
-		t = tokens->next;
-		mosquitto__free(tokens->topic);
-		mosquitto__free(tokens);
-		tokens = t;
-
-		sharename = tokens->topic;
-
-		tokens->topic = mosquitto__strdup("");
-		if(!tokens->topic){
-			tokens->topic = sharename;
-			sub__topic_tokens_free(tokens);
-			return MOSQ_ERR_PROTOCOL;
-		}
-		tokens->topic_len = 0;
-	}
-
-	HASH_FIND(hh, root, tokens->topic, tokens->topic_len, subhier);
+	HASH_FIND(hh, root, topics[0], strlen(topics[0]), subhier);
 	if(subhier){
 		*reason = MQTT_RC_NO_SUBSCRIPTION_EXISTED;
-		rc = sub__remove_recurse(db, context, subhier, tokens, reason, sharename);
+		rc = sub__remove_recurse(context, subhier, topics, reason, sharename);
 	}
 
-	sub__topic_tokens_free(tokens);
+	mosquitto__free(local_sub);
+	mosquitto__free(topics);
 
 	return rc;
 }
 
-int sub__messages_queue(struct mosquitto_db *db, const char *source_id, const char *topic, int qos, int retain, struct mosquitto_msg_store **stored)
+int sub__messages_queue(const char *source_id, const char *topic, uint8_t qos, int retain, struct mosquitto_msg_store **stored)
 {
-	int rc = 0;
+	int rc = MOSQ_ERR_SUCCESS, rc2;
 	struct mosquitto__subhier *subhier;
-	struct sub__token *tokens = NULL;
+	char **split_topics = NULL;
+	char *local_topic = NULL;
 
-	assert(db);
 	assert(topic);
 
-	if(sub__topic_tokenise(topic, &tokens)) return 1;
+	if(sub__topic_tokenise(topic, &local_topic, &split_topics, NULL)) return 1;
 
 	/* Protect this message until we have sent it to all
 	clients - this is required because websockets client calls
@@ -818,20 +650,20 @@ int sub__messages_queue(struct mosquitto_db *db, const char *source_id, const ch
 	*/
 	db__msg_store_ref_inc(*stored);
 
-	HASH_FIND(hh, db->subs, tokens->topic, tokens->topic_len, subhier);
+	HASH_FIND(hh, db.subs, split_topics[0], strlen(split_topics[0]), subhier);
 	if(subhier){
-		if(retain){
-			/* We have a message that needs to be retained, so ensure that the subscription
-			 * tree for its topic exists.
-			 */
-			sub__add_context(db, NULL, 0, 0, 0, subhier, tokens, NULL);
-		}
-		rc = sub__search(db, subhier, tokens, source_id, topic, qos, retain, *stored, true);
+		rc = sub__search(subhier, split_topics, source_id, topic, qos, retain, *stored);
 	}
-	sub__topic_tokens_free(tokens);
 
+	if(retain){
+		rc2 = retain__store(topic, *stored, split_topics);
+		if(rc2) rc = rc2;
+	}
+
+	mosquitto__free(split_topics);
+	mosquitto__free(local_topic);
 	/* Remove our reference and free if needed. */
-	db__msg_store_ref_dec(db, stored);
+	db__msg_store_ref_dec(stored);
 
 	return rc;
 }
@@ -846,7 +678,7 @@ static struct mosquitto__subhier *tmp_remove_subs(struct mosquitto__subhier *sub
 		return NULL;
 	}
 
-	if(sub->children || sub->subs || sub->retained){
+	if(sub->children || sub->subs){
 		return NULL;
 	}
 
@@ -857,7 +689,6 @@ static struct mosquitto__subhier *tmp_remove_subs(struct mosquitto__subhier *sub
 
 	if(parent->subs == NULL
 			&& parent->children == NULL
-			&& parent->retained == NULL
 			&& parent->shared == NULL
 			&& parent->parent){
 
@@ -868,7 +699,7 @@ static struct mosquitto__subhier *tmp_remove_subs(struct mosquitto__subhier *sub
 }
 
 
-static int sub__clean_session_shared(struct mosquitto_db *db, struct mosquitto *context)
+static int sub__clean_session_shared(struct mosquitto *context)
 {
 	int i;
 	struct mosquitto__subleaf *leaf;
@@ -882,7 +713,7 @@ static int sub__clean_session_shared(struct mosquitto_db *db, struct mosquitto *
 		while(leaf){
 			if(leaf->context==context){
 #ifdef WITH_SYS_TREE
-				db->shared_subscription_count--;
+				db.shared_subscription_count--;
 #endif
 				sub__remove_shared_leaf(context->shared_subs[i]->hier, context->shared_subs[i]->shared, leaf);
 				break;
@@ -891,7 +722,6 @@ static int sub__clean_session_shared(struct mosquitto_db *db, struct mosquitto *
 		}
 		if(context->shared_subs[i]->hier->subs == NULL
 				&& context->shared_subs[i]->hier->children == NULL
-				&& context->shared_subs[i]->hier->retained == NULL
 				&& context->shared_subs[i]->hier->shared == NULL
 				&& context->shared_subs[i]->hier->parent){
 
@@ -912,7 +742,7 @@ static int sub__clean_session_shared(struct mosquitto_db *db, struct mosquitto *
 
 /* Remove all subscriptions for a client.
  */
-int sub__clean_session(struct mosquitto_db *db, struct mosquitto *context)
+int sub__clean_session(struct mosquitto *context)
 {
 	int i;
 	struct mosquitto__subleaf *leaf;
@@ -926,7 +756,7 @@ int sub__clean_session(struct mosquitto_db *db, struct mosquitto *context)
 		while(leaf){
 			if(leaf->context==context){
 #ifdef WITH_SYS_TREE
-				db->subscription_count--;
+				db.subscription_count--;
 #endif
 				DL_DELETE(context->subs[i]->subs, leaf);
 				mosquitto__free(leaf);
@@ -936,7 +766,6 @@ int sub__clean_session(struct mosquitto_db *db, struct mosquitto *context)
 		}
 		if(context->subs[i]->subs == NULL
 				&& context->subs[i]->children == NULL
-				&& context->subs[i]->retained == NULL
 				&& context->subs[i]->shared == NULL
 				&& context->subs[i]->parent){
 
@@ -951,7 +780,7 @@ int sub__clean_session(struct mosquitto_db *db, struct mosquitto *context)
 	context->subs = NULL;
 	context->sub_count = 0;
 
-	return sub__clean_session_shared(db, context);
+	return sub__clean_session_shared(context);
 }
 
 void sub__tree_print(struct mosquitto__subhier *root, int level)
@@ -975,164 +804,9 @@ void sub__tree_print(struct mosquitto__subhier *root, int level)
 			}
 			leaf = leaf->next;
 		}
-		if(branch->retained){
-			printf(" (r)");
-		}
 		printf("\n");
 	}
 
 		sub__tree_print(branch->children, level+1);
 	}
 }
-
-static int retain__process(struct mosquitto_db *db, struct mosquitto__subhier *branch, struct mosquitto *context, int sub_qos, uint32_t subscription_identifier, time_t now)
-{
-	int rc = 0;
-	int qos;
-	uint16_t mid;
-	mosquitto_property *properties = NULL;
-	struct mosquitto_msg_store *retained;
-
-	if(branch->retained->message_expiry_time > 0 && now >= branch->retained->message_expiry_time){
-		db__msg_store_ref_dec(db, &branch->retained);
-		branch->retained = NULL;
-#ifdef WITH_SYS_TREE
-		db->retained_count--;
-#endif
-		return MOSQ_ERR_SUCCESS;
-	}
-
-	retained = branch->retained;
-
-	rc = mosquitto_acl_check(db, context, retained->topic, retained->payloadlen, UHPA_ACCESS(retained->payload, retained->payloadlen),
-			retained->qos, retained->retain, MOSQ_ACL_READ);
-	if(rc == MOSQ_ERR_ACL_DENIED){
-		return MOSQ_ERR_SUCCESS;
-	}else if(rc != MOSQ_ERR_SUCCESS){
-		return rc;
-	}
-
-	/* Check for original source access */
-	if(db->config->check_retain_source && retained->origin != mosq_mo_broker && retained->source_id){
-		struct mosquitto retain_ctxt;
-		memset(&retain_ctxt, 0, sizeof(struct mosquitto));
-
-		retain_ctxt.id = retained->source_id;
-		retain_ctxt.username = retained->source_username;
-		retain_ctxt.listener = retained->source_listener;
-
-		rc = acl__find_acls(db, &retain_ctxt);
-		if(rc) return rc;
-
-		rc = mosquitto_acl_check(db, &retain_ctxt, retained->topic, retained->payloadlen, UHPA_ACCESS(retained->payload, retained->payloadlen),
-				retained->qos, retained->retain, MOSQ_ACL_WRITE);
-		if(rc == MOSQ_ERR_ACL_DENIED){
-			return MOSQ_ERR_SUCCESS;
-		}else if(rc != MOSQ_ERR_SUCCESS){
-			return rc;
-		}
-	}
-
-	if (db->config->upgrade_outgoing_qos){
-		qos = sub_qos;
-	} else {
-		qos = retained->qos;
-		if(qos > sub_qos) qos = sub_qos;
-	}
-	if(qos > 0){
-		mid = mosquitto__mid_generate(context);
-	}else{
-		mid = 0;
-	}
-	if(subscription_identifier > 0){
-		mosquitto_property_add_varint(&properties, MQTT_PROP_SUBSCRIPTION_IDENTIFIER, subscription_identifier);
-	}
-	return db__message_insert(db, context, mid, mosq_md_out, qos, true, retained, properties);
-}
-
-static int retain__search(struct mosquitto_db *db, struct mosquitto__subhier *subhier, struct sub__token *tokens, struct mosquitto *context, const char *sub, int sub_qos, uint32_t subscription_identifier, time_t now, int level)
-{
-	struct mosquitto__subhier *branch, *branch_tmp;
-	int flag = 0;
-
-	if(!strcmp(tokens->topic, "#") && !tokens->next){
-		HASH_ITER(hh, subhier->children, branch, branch_tmp){
-			/* Set flag to indicate that we should check for retained messages
-			 * on "foo" when we are subscribing to e.g. "foo/#" and then exit
-			 * this function and return to an earlier retain__search().
-			 */
-			flag = -1;
-			if(branch->retained){
-				retain__process(db, branch, context, sub_qos, subscription_identifier, now);
-			}
-			if(branch->children){
-				retain__search(db, branch, tokens, context, sub, sub_qos, subscription_identifier, now, level+1);
-			}
-		}
-	}else{
-		if(!strcmp(tokens->topic, "+")){
-			HASH_ITER(hh, subhier->children, branch, branch_tmp){
-				if(tokens->next){
-					if(retain__search(db, branch, tokens->next, context, sub, sub_qos, subscription_identifier, now, level+1) == -1
-							|| (tokens->next && !strcmp(tokens->next->topic, "#") && level>0)){
-
-						if(branch->retained){
-							retain__process(db, branch, context, sub_qos, subscription_identifier, now);
-						}
-					}
-				}else{
-					if(branch->retained){
-						retain__process(db, branch, context, sub_qos, subscription_identifier, now);
-					}
-				}
-			}
-		}else{
-			HASH_FIND(hh, subhier->children, tokens->topic, tokens->topic_len, branch);
-			if(branch){
-				if(tokens->next){
-					if(retain__search(db, branch, tokens->next, context, sub, sub_qos, subscription_identifier, now, level+1) == -1
-							|| (tokens->next && !strcmp(tokens->next->topic, "#") && level>0)){
-
-						if(branch->retained){
-							retain__process(db, branch, context, sub_qos, subscription_identifier, now);
-						}
-					}
-				}else{
-					if(branch->retained){
-						retain__process(db, branch, context, sub_qos, subscription_identifier, now);
-					}
-				}
-			}
-		}
-	}
-	return flag;
-}
-
-int sub__retain_queue(struct mosquitto_db *db, struct mosquitto *context, const char *sub, int sub_qos, uint32_t subscription_identifier)
-{
-	struct mosquitto__subhier *subhier;
-	struct sub__token *tokens = NULL, *tail;
-	time_t now;
-
-	assert(db);
-	assert(context);
-	assert(sub);
-
-	if(sub__topic_tokenise(sub, &tokens)) return 1;
-
-	HASH_FIND(hh, db->subs, tokens->topic, tokens->topic_len, subhier);
-
-	if(subhier){
-		now = time(NULL);
-		retain__search(db, subhier, tokens, context, sub, sub_qos, subscription_identifier, now, 0);
-	}
-	while(tokens){
-		tail = tokens->next;
-		mosquitto__free(tokens->topic);
-		mosquitto__free(tokens);
-		tokens = tail;
-	}
-
-	return MOSQ_ERR_SUCCESS;
-}
-
