@@ -1,30 +1,19 @@
 /*
-Copyright (c) 2014-2020 Roger Light <roger@atchoo.org>
-All rights reserved.
+Copyright (c) 2014-2019 Roger Light <roger@atchoo.org>
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+All rights reserved. This program and the accompanying materials
+are made available under the terms of the Eclipse Public License 2.0
+and Eclipse Distribution License v1.0 which accompany this distribution.
 
-1. Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
-2. Redistributions in binary form must reproduce the above copyright
-   notice, this list of conditions and the following disclaimer in the
-   documentation and/or other materials provided with the distribution.
-3. Neither the name of mosquitto nor the names of its
-   contributors may be used to endorse or promote products derived from
-   this software without specific prior written permission.
+The Eclipse Public License is available at
+   https://www.eclipse.org/legal/epl-2.0/
+and the Eclipse Distribution License is available at
+  http://www.eclipse.org/org/documents/edl-v10.php.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-POSSIBILITY OF SUCH DAMAGE.
+SPDX-License-Identifier: EPL-2.0 OR EDL-1.0
+
+Contributors:
+   Roger Light - initial implementation and documentation.
 */
 
 #ifdef WITH_WEBSOCKETS
@@ -54,26 +43,16 @@ POSSIBILITY OF SUCH DAMAGE.
 #define WS_SERV_BUF_SIZE 4096
 #define WS_TX_BUF_SIZE (WS_SERV_BUF_SIZE*2)
 
-extern struct mosquitto_db int_db;
-
-#if defined(LWS_LIBRARY_VERSION_NUMBER)
 static int callback_mqtt(
-#else
-static int callback_mqtt(struct libwebsocket_context *context,
-#endif
-		struct libwebsocket *wsi,
-		enum libwebsocket_callback_reasons reason,
+		struct lws *wsi,
+		enum lws_callback_reasons reason,
 		void *user,
 		void *in,
 		size_t len);
 
-#if defined(LWS_LIBRARY_VERSION_NUMBER)
 static int callback_http(
-#else
-static int callback_http(struct libwebsocket_context *context,
-#endif
-	struct libwebsocket *wsi,
-	enum libwebsocket_callback_reasons reason,
+	struct lws *wsi,
+	enum lws_callback_reasons reason,
 	void *user,
 	void *in,
 	size_t len);
@@ -88,79 +67,43 @@ struct libws_http_data {
 	FILE *fptr;
 };
 
-static struct libwebsocket_protocols protocols[] = {
+static struct lws_protocols protocols[] = {
 	/* first protocol must always be HTTP handler */
 	{
 		"http-only",						/* name */
 		callback_http,						/* lws_callback_function */
 		sizeof (struct libws_http_data),	/* per_session_data_size */
 		0,									/* rx_buffer_size */
-#ifndef LWS_LIBRARY_VERSION_NUMBER
-		0,									/* no_buffer_all_partial_tx v1.3 only */
-#endif
-#ifdef LWS_FEATURE_PROTOCOLS_HAS_ID_FIELD
 		0,									/* id */
-#endif
-#ifdef LWS_LIBRARY_VERSION_NUMBER
 		NULL,								/* user v1.4 on */
-#  if LWS_LIBRARY_VERSION_NUMBER >= 2003000
 		WS_TX_BUF_SIZE						/* tx_packet_size v2.3.0 */
-#  endif
-#endif
 	},
 	{
 		"mqtt",
 		callback_mqtt,
 		sizeof(struct libws_mqtt_data),
 		0,									/* rx_buffer_size */
-#ifndef LWS_LIBRARY_VERSION_NUMBER
-		0,									/* no_buffer_all_partial_tx v1.3 only */
-#endif
-#ifdef LWS_FEATURE_PROTOCOLS_HAS_ID_FIELD
 		1,									/* id */
-#endif
-#ifdef LWS_LIBRARY_VERSION_NUMBER
 		NULL,								/* user v1.4 on */
-#  if LWS_LIBRARY_VERSION_NUMBER >= 2003000
 		WS_TX_BUF_SIZE						/* tx_packet_size v2.3.0 */
-#  endif
-#endif
 	},
 	{
 		"mqttv3.1",
 		callback_mqtt,
 		sizeof(struct libws_mqtt_data),
 		0,									/* rx_buffer_size */
-#ifndef LWS_LIBRARY_VERSION_NUMBER
-		0,									/* no_buffer_all_partial_tx v1.3 only */
-#endif
-#ifdef LWS_FEATURE_PROTOCOLS_HAS_ID_FIELD
 		2,									/* id */
-#endif
-#ifdef LWS_LIBRARY_VERSION_NUMBER
 		NULL,								/* user v1.4 on */
-#  if LWS_LIBRARY_VERSION_NUMBER >= 2003000
 		WS_TX_BUF_SIZE						/* tx_packet_size v2.3.0 */
-#  endif
-#endif
 	},
 	{
 		NULL,
 		NULL,
 		0,
 		0,									/* rx_buffer_size */
-#ifndef LWS_LIBRARY_VERSION_NUMBER
-		0,									/* no_buffer_all_partial_tx v1.3 only */
-#endif
-#ifdef LWS_FEATURE_PROTOCOLS_HAS_ID_FIELD
 		0,									/* id */
-#endif
-#ifdef LWS_LIBRARY_VERSION_NUMBER
 		NULL,								/* user v1.4 on */
-#  if LWS_LIBRARY_VERSION_NUMBER >= 2003000
 		0									/* tx_packet_size v2.3.0 */
-#  endif
-#endif
 	}
 };
 
@@ -168,49 +111,40 @@ static void easy_address(int sock, struct mosquitto *mosq)
 {
 	char address[1024];
 
-	if(!net__socket_get_address(sock, address, 1024)){
+	if(!net__socket_get_address(sock, address, 1024, &mosq->remote_port)){
 		mosq->address = mosquitto__strdup(address);
 	}
 }
 
-#if defined(LWS_LIBRARY_VERSION_NUMBER)
 static int callback_mqtt(
-#else
-static int callback_mqtt(struct libwebsocket_context *context,
-#endif
-		struct libwebsocket *wsi,
-		enum libwebsocket_callback_reasons reason,
+		struct lws *wsi,
+		enum lws_callback_reasons reason,
 		void *user,
 		void *in,
 		size_t len)
 {
-	struct mosquitto_db *db;
 	struct mosquitto *mosq = NULL;
 	struct mosquitto__packet *packet;
 	size_t txlen;
 	int count;
-	const struct libwebsocket_protocols *p;
+	unsigned int ucount;
+	const struct lws_protocols *p;
 	struct libws_mqtt_data *u = (struct libws_mqtt_data *)user;
 	size_t pos;
 	uint8_t *buf;
 	int rc;
 	uint8_t byte;
 
-	db = &int_db;
-
 	switch (reason) {
 		case LWS_CALLBACK_ESTABLISHED:
-			mosq = context__init(db, WEBSOCKET_CLIENT);
+			mosq = context__init(WEBSOCKET_CLIENT);
 			if(mosq){
-				p = libwebsockets_get_protocol(wsi);
+				p = lws_get_protocol(wsi);
 				mosq->listener = p->user;
 				if(!mosq->listener){
 					mosquitto__free(mosq);
 					return -1;
 				}
-#if !defined(LWS_LIBRARY_VERSION_NUMBER)
-				mosq->ws_context = context;
-#endif
 				mosq->wsi = wsi;
 #ifdef WITH_TLS
 				if(in){
@@ -224,7 +158,7 @@ static int callback_mqtt(struct libwebsocket_context *context,
 			}else{
 				return -1;
 			}
-			easy_address(libwebsocket_get_socket_fd(wsi), mosq);
+			easy_address(lws_get_socket_fd(wsi), mosq);
 			if(!mosq->address){
 				/* getpeername and inet_ntop failed and not a bridge */
 				mosquitto__free(mosq);
@@ -232,7 +166,7 @@ static int callback_mqtt(struct libwebsocket_context *context,
 				return -1;
 			}
 			if(mosq->listener->max_connections > 0 && mosq->listener->client_count > mosq->listener->max_connections){
-				if(db->config->connection_messages == true){
+				if(db.config->connection_messages == true){
 					log__printf(NULL, MOSQ_LOG_NOTICE, "Client connection from %s denied: max_connections exceeded.", mosq->address);
 				}
 				mosquitto__free(mosq->address);
@@ -240,8 +174,9 @@ static int callback_mqtt(struct libwebsocket_context *context,
 				u->mosq = NULL;
 				return -1;
 			}
-			mosq->sock = libwebsocket_get_socket_fd(wsi);
-			HASH_ADD(hh_sock, db->contexts_by_sock, sock, sizeof(mosq->sock), mosq);
+			mosq->sock = lws_get_socket_fd(wsi);
+			HASH_ADD(hh_sock, db.contexts_by_sock, sock, sizeof(mosq->sock), mosq);
+			mux__add_in(mosq);
 			break;
 
 		case LWS_CALLBACK_CLOSED:
@@ -251,15 +186,16 @@ static int callback_mqtt(struct libwebsocket_context *context,
 			mosq = u->mosq;
 			if(mosq){
 				if(mosq->sock != INVALID_SOCKET){
-					HASH_DELETE(hh_sock, db->contexts_by_sock, mosq);
+					HASH_DELETE(hh_sock, db.contexts_by_sock, mosq);
 					mosq->sock = INVALID_SOCKET;
 					mosq->pollfd_index = -1;
+					mux__delete(mosq);
 				}
 				mosq->wsi = NULL;
 #ifdef WITH_TLS
 				mosq->ssl = NULL;
 #endif
-				do_disconnect(db, mosq, MOSQ_ERR_CONN_LOST);
+				do_disconnect(mosq, MOSQ_ERR_CONN_LOST);
 			}
 			break;
 
@@ -272,7 +208,10 @@ static int callback_mqtt(struct libwebsocket_context *context,
 				return -1;
 			}
 
-			db__message_write(db, mosq);
+			rc = db__message_write_inflight_out_latest(mosq);
+			if(rc) return -1;
+			rc = db__message_write_queued_out(mosq);
+			if(rc) return -1;
 
 			if(mosq->out_packet && !mosq->current_out_packet){
 				mosq->current_out_packet = mosq->out_packet;
@@ -282,25 +221,25 @@ static int callback_mqtt(struct libwebsocket_context *context,
 				}
 			}
 
-			if(mosq->current_out_packet && !lws_send_pipe_choked(mosq->wsi)){
+			while(mosq->current_out_packet && !lws_send_pipe_choked(mosq->wsi)){
 				packet = mosq->current_out_packet;
 
 				if(packet->pos == 0 && packet->to_process == packet->packet_length){
 					/* First time this packet has been dealt with.
 					 * libwebsockets requires that the payload has
-					 * LWS_SEND_BUFFER_PRE_PADDING space available before the
-					 * actual data and LWS_SEND_BUFFER_POST_PADDING afterwards.
+					 * LWS_PRE space available before the
+					 * actual data.
 					 * We've already made the payload big enough to allow this,
 					 * but need to move it into position here. */
-					memmove(&packet->payload[LWS_SEND_BUFFER_PRE_PADDING], packet->payload, packet->packet_length);
-					packet->pos += LWS_SEND_BUFFER_PRE_PADDING;
+					memmove(&packet->payload[LWS_PRE], packet->payload, packet->packet_length);
+					packet->pos += LWS_PRE;
 				}
 				if(packet->to_process > WS_TX_BUF_SIZE){
 					txlen = WS_TX_BUF_SIZE;
 				}else{
 					txlen = packet->to_process;
 				}
-				count = libwebsocket_write(wsi, &packet->payload[packet->pos], txlen, LWS_WRITE_BINARY);
+				count = lws_write(wsi, &packet->payload[packet->pos], txlen, LWS_WRITE_BINARY);
 				if(count < 0){
 					if (mosq->state == mosq_cs_disconnect_ws
 							|| mosq->state == mosq_cs_disconnecting
@@ -310,11 +249,12 @@ static int callback_mqtt(struct libwebsocket_context *context,
 					}
 					return 0;
 				}
+				ucount = (unsigned int)count;
 #ifdef WITH_SYS_TREE
-				g_bytes_sent += count;
+				g_bytes_sent += ucount;
 #endif
-				packet->to_process -= count;
-				packet->pos += count;
+				packet->to_process -= ucount;
+				packet->pos += ucount;
 				if(packet->to_process > 0){
 					if (mosq->state == mosq_cs_disconnect_ws
 							|| mosq->state == mosq_cs_disconnecting
@@ -344,7 +284,7 @@ static int callback_mqtt(struct libwebsocket_context *context,
 				packet__cleanup(packet);
 				mosquitto__free(packet);
 
-				mosq->next_msg_out = mosquitto_time() + mosq->keepalive;
+				mosq->next_msg_out = db.now_s + mosq->keepalive;
 			}
 			if (mosq->state == mosq_cs_disconnect_ws
 					|| mosq->state == mosq_cs_disconnecting
@@ -353,7 +293,7 @@ static int callback_mqtt(struct libwebsocket_context *context,
 				return -1;
 			}
 			if(mosq->current_out_packet){
-				libwebsocket_callback_on_writable(mosq->ws_context, mosq->wsi);
+				lws_callback_on_writable(mosq->wsi);
 			}
 			break;
 
@@ -393,7 +333,7 @@ static int callback_mqtt(struct libwebsocket_context *context,
 						mosq->in_packet.remaining_length += (byte & 127) * mosq->in_packet.remaining_mult;
 						mosq->in_packet.remaining_mult *= 128;
 					}while((byte & 128) != 0);
-					mosq->in_packet.remaining_count *= -1;
+					mosq->in_packet.remaining_count = (int8_t)(mosq->in_packet.remaining_count -1);
 
 					if(mosq->in_packet.remaining_length > 0){
 						mosq->in_packet.payload = mosquitto__malloc(mosq->in_packet.remaining_length*sizeof(uint8_t));
@@ -404,15 +344,15 @@ static int callback_mqtt(struct libwebsocket_context *context,
 					}
 				}
 				if(mosq->in_packet.to_process>0){
-					if(len - pos >= mosq->in_packet.to_process){
+					if((uint32_t)len - pos >= mosq->in_packet.to_process){
 						memcpy(&mosq->in_packet.payload[mosq->in_packet.pos], &buf[pos], mosq->in_packet.to_process);
 						mosq->in_packet.pos += mosq->in_packet.to_process;
 						pos += mosq->in_packet.to_process;
 						mosq->in_packet.to_process = 0;
 					}else{
 						memcpy(&mosq->in_packet.payload[mosq->in_packet.pos], &buf[pos], len-pos);
-						mosq->in_packet.pos += len-pos;
-						mosq->in_packet.to_process -= len-pos;
+						mosq->in_packet.pos += (uint32_t)(len-pos);
+						mosq->in_packet.to_process -= (uint32_t)(len-pos);
 						return 0;
 					}
 				}
@@ -425,20 +365,20 @@ static int callback_mqtt(struct libwebsocket_context *context,
 					G_PUB_MSGS_RECEIVED_INC(1);
 				}
 #endif
-				rc = handle__packet(db, mosq);
+				rc = handle__packet(mosq);
 
 				/* Free data and reset values */
 				packet__cleanup(&mosq->in_packet);
 
-				mosq->last_msg_in = mosquitto_time();
+				keepalive__update(mosq);
 
 				if(rc && (mosq->out_packet || mosq->current_out_packet)) {
 					if(mosq->state != mosq_cs_disconnecting){
 						mosquitto__set_state(mosq, mosq_cs_disconnect_ws);
 					}
-					libwebsocket_callback_on_writable(mosq->ws_context, mosq->wsi);
+					lws_callback_on_writable(mosq->wsi);
 				} else if (rc) {
-					do_disconnect(db, mosq, MOSQ_ERR_CONN_LOST);
+					do_disconnect(mosq, MOSQ_ERR_CONN_LOST);
 					return -1;
 				}
 			}
@@ -453,10 +393,7 @@ static int callback_mqtt(struct libwebsocket_context *context,
 
 
 static char *http__canonical_filename(
-#ifndef LWS_LIBRARY_VERSION_NUMBER
-		struct libwebsocket_context *context,
-#endif
-		struct libwebsocket *wsi,
+		struct lws *wsi,
 		const char *in,
 		const char *http_dir)
 {
@@ -471,7 +408,7 @@ static char *http__canonical_filename(
 	}
 	filename = mosquitto__malloc(slen);
 	if(!filename){
-		libwebsockets_return_http_status(context, wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
+		lws_return_http_status(wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
 		return NULL;
 	}
 	if(((char *)in)[inlen-1] == '/'){
@@ -486,7 +423,7 @@ static char *http__canonical_filename(
 	filename_canonical = _fullpath(NULL, filename, 0);
 	mosquitto__free(filename);
 	if(!filename_canonical){
-		libwebsockets_return_http_status(context, wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
+		lws_return_http_status(wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
 		return NULL;
 	}
 #else
@@ -494,13 +431,13 @@ static char *http__canonical_filename(
 	mosquitto__free(filename);
 	if(!filename_canonical){
 		if(errno == EACCES){
-			libwebsockets_return_http_status(context, wsi, HTTP_STATUS_FORBIDDEN, NULL);
+			lws_return_http_status(wsi, HTTP_STATUS_FORBIDDEN, NULL);
 		}else if(errno == EINVAL || errno == EIO || errno == ELOOP){
-			libwebsockets_return_http_status(context, wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
+			lws_return_http_status(wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
 		}else if(errno == ENAMETOOLONG){
-			libwebsockets_return_http_status(context, wsi, HTTP_STATUS_REQ_URI_TOO_LONG, NULL);
+			lws_return_http_status(wsi, HTTP_STATUS_REQ_URI_TOO_LONG, NULL);
 		}else if(errno == ENOENT || errno == ENOTDIR){
-			libwebsockets_return_http_status(context, wsi, HTTP_STATUS_NOT_FOUND, NULL);
+			lws_return_http_status(wsi, HTTP_STATUS_NOT_FOUND, NULL);
 		}
 		return NULL;
 	}
@@ -508,7 +445,7 @@ static char *http__canonical_filename(
 	if(strncmp(http_dir, filename_canonical, strlen(http_dir))){
 		/* Requested file isn't within http_dir, deny access. */
 		free(filename_canonical);
-		libwebsockets_return_http_status(context, wsi, HTTP_STATUS_FORBIDDEN, NULL);
+		lws_return_http_status(wsi, HTTP_STATUS_FORBIDDEN, NULL);
 		return NULL;
 	}
 
@@ -516,13 +453,9 @@ static char *http__canonical_filename(
 }
 
 
-#if defined(LWS_LIBRARY_VERSION_NUMBER)
 static int callback_http(
-#else
-static int callback_http(struct libwebsocket_context *context,
-#endif
-		struct libwebsocket *wsi,
-		enum libwebsocket_callback_reasons reason,
+		struct lws *wsi,
+		enum lws_callback_reasons reason,
 		void *user,
 		void *in,
 		size_t len)
@@ -532,10 +465,10 @@ static int callback_http(struct libwebsocket_context *context,
 	char *http_dir;
 	size_t buflen;
 	size_t wlen;
+	int rc;
 	char *filename_canonical;
 	unsigned char buf[4096];
 	struct stat filestat;
-	struct mosquitto_db *db = &int_db;
 	struct mosquitto *mosq;
 	struct lws_pollargs *pollargs = (struct lws_pollargs *)in;
 
@@ -547,11 +480,7 @@ static int callback_http(struct libwebsocket_context *context,
 				return -1;
 			}
 
-#if defined(LWS_LIBRARY_VERSION_NUMBER)
 			hack = (struct libws_mqtt_hack *)lws_context_user(lws_get_context(wsi));
-#else
-			hack = (struct libws_mqtt_hack *)libwebsocket_context_user(context);
-#endif
 			if(!hack){
 				return -1;
 			}
@@ -564,26 +493,22 @@ static int callback_http(struct libwebsocket_context *context,
 
 			/* Forbid POST */
 			if(lws_hdr_total_length(wsi, WSI_TOKEN_POST_URI)){
-				libwebsockets_return_http_status(context, wsi, HTTP_STATUS_METHOD_NOT_ALLOWED, NULL);
+				lws_return_http_status(wsi, HTTP_STATUS_METHOD_NOT_ALLOWED, NULL);
 				return -1;
 			}
 
-#if defined(LWS_LIBRARY_VERSION_NUMBER)
 			filename_canonical = http__canonical_filename(wsi, (char *)in, http_dir);
-#else
-			filename_canonical = http__canonical_filename(context, wsi, (char *)in, http_dir);
-#endif
 			if(!filename_canonical) return -1;
 
 			u->fptr = fopen(filename_canonical, "rb");
 			if(!u->fptr){
 				free(filename_canonical);
-				libwebsockets_return_http_status(context, wsi, HTTP_STATUS_NOT_FOUND, NULL);
+				lws_return_http_status(wsi, HTTP_STATUS_NOT_FOUND, NULL);
 				return -1;
 			}
 			if(fstat(fileno(u->fptr), &filestat) < 0){
 				free(filename_canonical);
-				libwebsockets_return_http_status(context, wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
+				lws_return_http_status(wsi, HTTP_STATUS_INTERNAL_SERVER_ERROR, NULL);
 				fclose(u->fptr);
 				u->fptr = NULL;
 				return -1;
@@ -596,14 +521,14 @@ static int callback_http(struct libwebsocket_context *context,
 				free(filename_canonical);
 
 				/* FIXME - use header functions from lws 2.x */
-				buflen = snprintf((char *)buf, 4096, "HTTP/1.0 302 OK\r\n"
+				buflen = (size_t)snprintf((char *)buf, 4096, "HTTP/1.0 302 OK\r\n"
 												"Location: %s/\r\n\r\n",
 												(char *)in);
-				return libwebsocket_write(wsi, buf, buflen, LWS_WRITE_HTTP);
+				return lws_write(wsi, buf, buflen, LWS_WRITE_HTTP);
 			}
 
 			if((filestat.st_mode & S_IFREG) != S_IFREG){
-				libwebsockets_return_http_status(context, wsi, HTTP_STATUS_FORBIDDEN, NULL);
+				lws_return_http_status(wsi, HTTP_STATUS_FORBIDDEN, NULL);
 				fclose(u->fptr);
 				u->fptr = NULL;
 				free(filename_canonical);
@@ -613,16 +538,16 @@ static int callback_http(struct libwebsocket_context *context,
 			log__printf(NULL, MOSQ_LOG_DEBUG, "http serving file \"%s\".", filename_canonical);
 			free(filename_canonical);
 			/* FIXME - use header functions from lws 2.x */
-			buflen = snprintf((char *)buf, 4096, "HTTP/1.0 200 OK\r\n"
+			buflen = (size_t)snprintf((char *)buf, 4096, "HTTP/1.0 200 OK\r\n"
 												"Server: mosquitto\r\n"
 												"Content-Length: %u\r\n\r\n",
 												(unsigned int)filestat.st_size);
-            if(libwebsocket_write(wsi, buf, buflen, LWS_WRITE_HTTP) < 0){
+            if(lws_write(wsi, buf, buflen, LWS_WRITE_HTTP) < 0){
 				fclose(u->fptr);
 				u->fptr = NULL;
 				return -1;
 			}
-			libwebsocket_callback_on_writable(context, wsi);
+			lws_callback_on_writable(wsi);
 			break;
 
 		case LWS_CALLBACK_HTTP_BODY:
@@ -647,9 +572,13 @@ static int callback_http(struct libwebsocket_context *context,
 						u->fptr = NULL;
 						return -1;
 					}
-					wlen = libwebsocket_write(wsi, buf, buflen, LWS_WRITE_HTTP);
+					rc = lws_write(wsi, buf, buflen, LWS_WRITE_HTTP);
+					if(rc < 0){
+						return -1;
+					}
+					wlen = (size_t)rc;
 					if(wlen < buflen){
-						if(fseek(u->fptr, buflen-wlen, SEEK_CUR) < 0){
+						if(fseek(u->fptr, (long)(buflen-wlen), SEEK_CUR) < 0){
 							fclose(u->fptr);
 							u->fptr = NULL;
 							return -1;
@@ -661,7 +590,7 @@ static int callback_http(struct libwebsocket_context *context,
 						}
 					}
 				}while(u->fptr && !lws_send_pipe_choked(wsi));
-				libwebsocket_callback_on_writable(context, wsi);
+				lws_callback_on_writable(wsi);
 			}else{
 				return -1;
 			}
@@ -679,9 +608,14 @@ static int callback_http(struct libwebsocket_context *context,
 		case LWS_CALLBACK_ADD_POLL_FD:
 		case LWS_CALLBACK_DEL_POLL_FD:
 		case LWS_CALLBACK_CHANGE_MODE_POLL_FD:
-			HASH_FIND(hh_sock, db->contexts_by_sock, &pollargs->fd, sizeof(pollargs->fd), mosq);
-			if(mosq && (pollargs->events & POLLOUT)){
-				mosq->ws_want_write = true;
+			HASH_FIND(hh_sock, db.contexts_by_sock, &pollargs->fd, sizeof(pollargs->fd), mosq);
+			if(mosq){
+				if(pollargs->events & POLLOUT){
+					mux__add_out(mosq);
+					mosq->ws_want_write = true;
+				}else{
+					mux__remove_out(mosq);
+				}
 			}
 			break;
 
@@ -707,18 +641,18 @@ static void log_wrap(int level, const char *line)
 	log__printf(NULL, MOSQ_LOG_WEBSOCKETS, "%s", l);
 }
 
-struct libwebsocket_context *mosq_websockets_init(struct mosquitto__listener *listener, const struct mosquitto__config *conf)
+struct lws_context *mosq_websockets_init(struct mosquitto__listener *listener, const struct mosquitto__config *conf)
 {
 	struct lws_context_creation_info info;
-	struct libwebsocket_protocols *p;
-	int protocol_count;
+	struct lws_protocols *p;
+	size_t protocol_count;
 	int i;
 	struct libws_mqtt_hack *user;
 
 	/* Count valid protocols */
 	for(protocol_count=0; protocols[protocol_count].name; protocol_count++);
 
-	p = mosquitto__calloc(protocol_count+1, sizeof(struct libwebsocket_protocols));
+	p = mosquitto__calloc(protocol_count+1, sizeof(struct lws_protocols));
 	if(!p){
 		log__printf(NULL, MOSQ_LOG_ERR, "Out of memory.");
 		return NULL;
@@ -742,20 +676,19 @@ struct libwebsocket_context *mosq_websockets_init(struct mosquitto__listener *li
 	info.ssl_cert_filepath = listener->certfile;
 	info.ssl_private_key_filepath = listener->keyfile;
 	info.ssl_cipher_list = listener->ciphers;
+#if defined(WITH_WEBSOCKETS) && LWS_LIBRARY_VERSION_NUMBER>=3001000
+	info.tls1_3_plus_cipher_list = listener->ciphers_tls13;
+#endif
 	if(listener->require_certificate){
 		info.options |= LWS_SERVER_OPTION_REQUIRE_VALID_OPENSSL_CLIENT_CERT;
 	}
 #endif
 
-#if LWS_LIBRARY_VERSION_MAJOR>1
 	info.options |= LWS_SERVER_OPTION_DO_SSL_GLOBAL_INIT;
-#endif
 	if(listener->socket_domain == AF_INET){
 		info.options |= LWS_SERVER_OPTION_DISABLE_IPV6;
 	}
-#if defined(LWS_LIBRARY_VERSION_NUMBER) && LWS_LIBRARY_VERSION_NUMBER>=1007000
     info.max_http_header_data = conf->websockets_headers_size;
-#endif
 
 	user = mosquitto__calloc(1, sizeof(struct libws_mqtt_hack));
 	if(!user){
@@ -779,15 +712,13 @@ struct libwebsocket_context *mosq_websockets_init(struct mosquitto__listener *li
 	}
 
 	info.user = user;
-#if defined(LWS_LIBRARY_VERSION_NUMBER) && LWS_LIBRARY_VERSION_NUMBER>=2004000
 	info.pt_serv_buf_size = WS_SERV_BUF_SIZE;
-#endif
 	listener->ws_protocol = p;
 
 	lws_set_log_level(conf->websockets_log_level, log_wrap);
 
 	log__printf(NULL, MOSQ_LOG_INFO, "Opening websockets listen socket on port %d.", listener->port);
-	return libwebsocket_create_context(&info);
+	return lws_create_context(&info);
 }
 
 
