@@ -5,25 +5,22 @@ import ptest
 
 tests = [
     #(ports required, 'path'),
-    (1, './01-connect-anon-denied.py'),
+    (1, './01-connect-allow-anonymous.py'),
     (1, './01-connect-bad-packet.py'),
     (1, './01-connect-disconnect-v5.py'),
-    (1, './01-connect-duplicate-v5.py'),
     (1, './01-connect-duplicate.py'),
-    (1, './01-connect-invalid-id-0-311.py'),
     (1, './01-connect-invalid-id-0.py'),
     (1, './01-connect-invalid-id-missing.py'),
     (1, './01-connect-invalid-id-utf8.py'),
     (1, './01-connect-invalid-protonum.py'),
     (1, './01-connect-invalid-reserved.py'),
-    (1, './01-connect-success-v5.py'),
     (1, './01-connect-success.py'),
     (1, './01-connect-uname-invalid-utf8.py'),
     (1, './01-connect-uname-no-flag.py'),
     (1, './01-connect-uname-no-password-denied.py'),
+    (1, './01-connect-uname-or-anon.py'),
     (1, './01-connect-uname-password-denied-no-will.py'),
     (1, './01-connect-uname-password-denied.py'),
-    (1, './01-connect-uname-password-success.py'),
     (1, './01-connect-uname-pwd-no-flag.py'),
     (2, './01-connect-zero-length-id.py'),
 
@@ -35,7 +32,6 @@ tests = [
     (1, './02-subpub-qos0-subscription-id.py'),
     (1, './02-subpub-qos0-topic-alias-unknown.py'),
     (1, './02-subpub-qos0-topic-alias.py'),
-    (1, './02-subpub-qos0-v5.py'),
     (1, './02-subpub-qos0.py'),
     (1, './02-subpub-qos1-bad-pubcomp.py'),
     (1, './02-subpub-qos1-bad-pubrec.py'),
@@ -43,7 +39,6 @@ tests = [
     (1, './02-subpub-qos1-message-expiry-will.py'),
     (1, './02-subpub-qos1-message-expiry.py'),
     (1, './02-subpub-qos1-nolocal.py'),
-    (1, './02-subpub-qos1-v5.py'),
     (1, './02-subpub-qos1.py'),
     (1, './02-subpub-qos2-1322.py'),
     (1, './02-subpub-qos2-bad-puback-1.py'),
@@ -52,7 +47,6 @@ tests = [
     (1, './02-subpub-qos2-pubrec-error.py'),
     (1, './02-subpub-qos2-receive-maximum-1.py'),
     (1, './02-subpub-qos2-receive-maximum-2.py'),
-    (1, './02-subpub-qos2-v5.py'),
     (1, './02-subpub-qos2.py'),
     (1, './02-subscribe-dollar-v5.py'),
     (1, './02-subscribe-invalid-utf8.py'),
@@ -64,9 +58,7 @@ tests = [
     (1, './02-unsubscribe-invalid-no-topic.py'),
     (1, './02-unsubscribe-qos0.py'),
     (1, './02-unsubscribe-qos1.py'),
-    (1, './02-unsubscribe-qos2-multiple-v5.py'),
     (1, './02-unsubscribe-qos2-multiple.py'),
-    (1, './02-unsubscribe-qos2-v5.py'),
     (1, './02-unsubscribe-qos2.py'),
 
     #(1, './03-publish-qos1-queued-bytes.py'),
@@ -81,6 +73,8 @@ tests = [
     (1, './03-publish-dollar.py'),
     (1, './03-publish-invalid-utf8.py'),
     (1, './03-publish-long-topic.py'),
+    (1, './03-publish-qos1-max-inflight-expire.py'),
+    (1, './03-publish-qos1-max-inflight.py'),
     (1, './03-publish-qos1-no-subscribers-v5.py'),
     (1, './03-publish-qos1-retain-disabled.py'),
     (1, './03-publish-qos1.py'),
@@ -108,9 +102,11 @@ tests = [
     (2, './06-bridge-br2b-disconnect-qos1.py'),
     (2, './06-bridge-br2b-disconnect-qos2.py'),
     (2, './06-bridge-br2b-remapping.py'),
+    (2, './06-bridge-clean-session.py'),
     (2, './06-bridge-fail-persist-resend-qos1.py'),
     (2, './06-bridge-fail-persist-resend-qos2.py'),
     (1, './06-bridge-no-local.py'),
+    (2, './06-bridge-outgoing-retain.py'),
     (3, './06-bridge-per-listener-settings.py'),
     (2, './06-bridge-reconnect-local-out.py'),
 
@@ -127,6 +123,7 @@ tests = [
     (1, './07-will-properties.py'),
     (1, './07-will-qos0.py'),
     (1, './07-will-reconnect-1273.py'),
+    (1, './07-will-takeover.py'),
 
     (2, './08-ssl-bridge.py'),
     (2, './08-ssl-connect-cert-auth-crl.py'),
@@ -162,6 +159,7 @@ tests = [
     (1, './09-plugin-auth-unpwd-success.py'),
     (1, './09-plugin-auth-v2-unpwd-fail.py'),
     (1, './09-plugin-auth-v2-unpwd-success.py'),
+    (1, './09-plugin-publish.py'),
     (1, './09-pwfile-parse-invalid.py'),
 
     (2, './10-listener-mount-point.py'),
@@ -186,6 +184,26 @@ tests = [
     (1, './12-prop-subpub-content-type.py'),
     (1, './12-prop-subpub-payload-format.py'),
     (1, './12-prop-topic-alias-invalid.py'),
+
+    (1, './13-malformed-publish-v5.py'),
+    (1, './13-malformed-subscribe-v5.py'),
+    (1, './13-malformed-unsubscribe-v5.py'),
+
+    (1, './14-dynsec-acl.py'),
+    (1, './14-dynsec-anon-group.py'),
+    (1, './14-dynsec-auth.py'),
+    (1, './14-dynsec-client.py'),
+    (1, './14-dynsec-client-invalid.py'),
+    (1, './14-dynsec-default-access.py'),
+    (1, './14-dynsec-disable-client.py'),
+    (1, './14-dynsec-group.py'),
+    (1, './14-dynsec-group-invalid.py'),
+    (1, './14-dynsec-modify-client.py'),
+    (1, './14-dynsec-modify-group.py'),
+    (1, './14-dynsec-modify-role.py'),
+    (1, './14-dynsec-plugin-invalid.py'),
+    (1, './14-dynsec-role.py'),
+    (1, './14-dynsec-role-invalid.py'),
     ]
 
 ptest.run_tests(tests)
