@@ -5,7 +5,7 @@ struct mosquitto_db{
 
 };
 
-int log__printf(struct mosquitto *mosq, int priority, const char *fmt, ...)
+int log__printf(struct mosquitto *mosq, unsigned int priority, const char *fmt, ...)
 {
 	return 0;
 }
